#!/usr/bin/python
"""
Datetime list
=====================

Copyright (C) 2005-2012 National Technical University of Athens

Copyright (C) 2005 Antonis Christofides, 2012 Stefanos Kozanis

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

from datetime import datetime
from pthelma.timeseries import (datetime_from_iso, isoformat_nosecs,
    _DT_BASE, _datetime_to_time_t, _time_t_to_datetime,)

from ctypes import CDLL, c_int, c_longlong,  c_char_p, byref, \
                   c_void_p

import platform
dickinson = CDLL('dickinson.dll' if platform.system()=='Windows'
                                                    else 'libdickinson.so')

dickinson.dl_get_item.restype = c_longlong
dickinson.dl_create.restype = c_void_p

class DatetimeList(list):

    def __init__(self, *args):
        self.dl_handle=None
        self.dl_handle = c_void_p(dickinson.dl_create())
        if self.dl_handle.value==0:
            raise MemoryError.Create('Could not allocate memory '
                                     'for date time list object')
        self.__dickinson = dickinson
        if len(args)>0:
            for item in args:
                self.append(item)

    def __del__(self):
        if self.dl_handle is None:
            return
        if self.dl_handle.value!=0:
            self.__dickinson.dl_free(self.dl_handle)
        self.dl_handle.value=0

    def append(self, x):
        d = datetime_from_iso(x) if (isinstance(x, unicode) or \
                                     isinstance(x, str)) else x
        if not isinstance(d, datetime):
            raise ValueError('Only date time items or ISO date strings '
                             'allowed for date time lists') 
        err_str_c = c_char_p()
        index_c = c_int()
        err_no_c = dickinson.dl_insert_record(self.dl_handle,
            c_longlong(_datetime_to_time_t(d)),
            byref(index_c), byref(err_str_c))
        if err_no_c!=0:
            raise Exception('Something wrong occured in dickinson '
                'function when setting a time series value. '+
                'Error message: '+repr(err_str_c.value))

    def insert(self, x):
        return self.append(x)

    def __repr__(self):
        return '[%s]'%(', '.join((repr(i) for i in self)),)

    def __str__(self):
        return '[%s]'%(', '.join((str(i) for i in self)),)

    def items(self):
        return [i for i in self]

    def __len__(self):
        return dickinson.dl_length(self.dl_handle)

    def __delitem__(self, key):
        if not isinstance(key, int):
            raise KeyError('Index is not an integer')
        if key<0 or key>self.__len__()-1:
            raise IndexError('Index out of bounds (%d)'%key)
        dickinson.dl_delete_item(self.dl_handle, c_int(key))

    def pop(self, *args):
        if self.__len__()<1:
            raise IndexError('Empty datetime list, cannot pop')
        key = args[0] if len(args)!=0 else self.__len__()-1
        if not isinstance(key, int):
            raise KeyError('Pop index is not an integer')
        if key<0 or key>self.__len__()-1:
            raise IndexError('Index out of bounds (%d)'%key)
        result = self.__getitem__(key)
        dickinson.dl_delete_item(self.dl_handle, c_int(key))
        return result

    def __contains__(self, item):
        index_c = dickinson.dl_get_i(self.dl_handle,
                                        c_longlong(_datetime_to_time_t(item)))
        return False if index_c<0 else True

    def __getitem__(self, key):
        if not isinstance(key, int):
            raise KeyError('Invalid key: %s'%(key,))
        if key<0 or key>self.__len__()-1:
            raise IndexError('Index out of bounds (%d)'%(key,))
        return _time_t_to_datetime(dickinson.dl_get_item(self.dl_handle,
               c_int(key)))

    def iterkeys(self):
        i = 0
        while i<self.__len__():
            yield self[i]
            i+=1
    __iter__ = iterkeys

    def __reversed__(self):
        i = self.__len__()-1
        while i>=0:
            yield self[i]
            i-=1

    def clear(self):
        i = self.__len__()-1
        while i>=0:
            dickinson.dl_delete_item(self.dl_handle, i)
            i-=1

    def index(self, x):
        d = datetime_from_iso(x) if (isinstance(x, unicode) or \
                                     isinstance(x, str)) else x
        if not isinstance(d, datetime):
            raise ValueError('Only date time items or ISO date strings '
                             'allowed for date time lists') 
        err_str_c = c_char_p()
        index_c = c_int()
        timestamp_c = c_longlong(_datetime_to_time_t(d))
        pos = dickinson.dl_get_i(self.dl_handle, timestamp_c)
        if pos<0:
            raise KeyError('No such item (%s) in date time list'%\
                           (x,))
        return pos

    def sort(self):
        pass

    def count(self, x):
        return 1 if x in self else 0

    def remove(self, x):
        pos = self.index(x)
        self.__delitem__(pos)



#!/usr/bin/python
"""
Tests for meteologger.

Copyright (C) 2009-2013 National Technical University of Athens
Copyright (C) 2013 TEI of Epirus

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

In order to run the meteologger tests, you must specify the
PTHELMA_TEST_METEOLOGGER variable to contain a json-formatted string of
parameters, like this:

    { "base_url": "http://localhost:8001/",
    "username": "admin",
    "password": "secret",
    "station_id": 1334,
    "variable_id": 1,
    "unit_of_measurement_id": 1,
    "time_zone_id": 1 }

The first three parameters are for connecting to an appropriate
database. Don't use a production database for that; although things
are normally cleaned up (e.g. test timeseries created are deleted), id
serial numbers will be affected and things might not be cleaned up if
there is an error.

The rest of the parameters are used when test timeseries are created.
"""

import cookielib
import json
from math import isnan
import os
import tempfile
from unittest import TestCase, skipUnless
from urllib2 import build_opener, HTTPCookieProcessor, Request, HTTPError

from pthelma.meteologger import Datafile_deltacom, Datafile_simple, \
    Datafile_wdat5, ConfigurationError
from pthelma.timeseries import Timeseries


def full_testdata_filename(filename):
    current_directory = os.path.dirname(os.path.abspath(__file__))
    full_pathname = os.path.join(current_directory, 'data', filename)
    return full_pathname


class RequestWithMethod(Request):
    """
    See http://benjamin.smedbergs.us/blog/2008-10-21/putting-and-deleteing
    """
    def __init__(self, method, *args, **kwargs):
        self._method = method
        Request.__init__(self, *args, **kwargs)

    def get_method(self):
        return self._method


def connect_to_server(base_url, username, password):
    cookiejar = cookielib.CookieJar()
    opener = build_opener(HTTPCookieProcessor(cookiejar))
    opener.open(base_url + 'accounts/login/').read()
    opener.addheaders = [('X-CSRFToken', cookie.value)
                         for cookie in cookiejar if cookie.name == 'csrftoken']
    data = 'username={0}&password={1}'.format(username, password)
    opener.open(base_url + 'accounts/login/', data)
    opener.addheaders = [('X-CSRFToken', cookie.value)
                         for cookie in cookiejar if cookie.name == 'csrftoken']
    return opener


def create_timeseries(opener, adict):
    # Create a timeseries on the server and return its id
    j = {'gentity': adict['station_id'],
         'variable': adict['variable_id'],
         'unit_of_measurement': adict['unit_of_measurement_id'],
         'time_zone': adict['time_zone_id'], }
    r = Request(adict['base_url'] + 'api/Timeseries/', data=json.dumps(j),
                headers={'Content-type': 'application/json'})
    fp = opener.open(r)
    response_text = fp.read()
    return json.loads(response_text)['id']


def delete_timeseries(opener, base_url, ts_id):
    url = base_url + 'api/Timeseries/{0}/'.format(ts_id)
    r = RequestWithMethod('DELETE', url)
    try:
        opener.open(r)
        # According to the theory at
        # http://stackoverflow.com/questions/7032890/ and elsewhere,
        # we shouldn't reach this point; in practice, however, I see
        # that a 204 response results in coming here, not in raising an
        # exception (and with no way to check the status code). I assume
        # that if we come here everything was OK.
    except HTTPError as e:
        # But just to be sure, also assume that a 204 response might
        # lead us here as well.
        if e.code != 204:
            raise


def read_timeseries(opener, base_url, ts):
    r = Request(base_url + 'api/tsdata/{0}/'.format(ts.id))
    fp = opener.open(r)
    ts.read(fp)


def get_server_from_env(adict):
    conn = os.getenv("PTHELMA_TEST_METEOLOGGER")
    v = json.loads(conn)
    for item in 'base_url username password station_id variable_id ' \
                'unit_of_measurement_id time_zone_id'.split():
        adict[item] = v[item]


@skipUnless(os.getenv('PTHELMA_TEST_METEOLOGGER'),
            "set PTHELMA_TEST_METEOLOGGER")
class _Test_logger(TestCase):
    class_being_tested = None
    datafilename = ''
    datafiledict = {}
    ref_ts1 = Timeseries(0)
    ref_ts1.read(open(full_testdata_filename('timeseries1.txt')))
    ref_ts2 = Timeseries(0)
    ref_ts2.read(open(full_testdata_filename('timeseries2.txt')))

    def setUp(self):
        get_server_from_env(self.__dict__)

        if not self.class_being_tested or not self.base_url:
            return

        # First 60 nonempty lines of test go to self.file1
        (fd, self.file1) = tempfile.mkstemp(text=True)
        fp = os.fdopen(fd, 'w')
        i = 0
        for line in open(full_testdata_filename(self.datafilename)):
            if i >= 60:
                break
            if not line.strip():
                continue
            i += 1
            fp.write(line)
        fp.close()

        # Connect to server
        self.opener = connect_to_server(self.base_url, self.username,
                                        self.password)

        # Create two timeseries
        self.timeseries_id1 = create_timeseries(self.opener, self.__dict__)
        self.timeseries_id2 = create_timeseries(self.opener, self.__dict__)
        self.datafile_fields = "0,{0},{1}".format(self.timeseries_id1,
                                                  self.timeseries_id2)
        self.ts1 = Timeseries(self.timeseries_id1)
        self.ts2 = Timeseries(self.timeseries_id2)

    def tearDown(self):
        if not self.class_being_tested or not self.base_url:
            return
        delete_timeseries(self.opener, self.base_url, self.timeseries_id2)
        delete_timeseries(self.opener, self.base_url, self.timeseries_id1)

    def check_config_test(self):
        self.assertRaises(ConfigurationError, self.class_being_tested,
                          self.base_url, self.opener, {})
        self.assertRaises(ConfigurationError, self.class_being_tested,
                          self.base_url, self.opener,
                          {'filename': 'hello', 'datafile_fields': '0',
                           'datafile_format': 'irrelevant',
                           'nonexistent_config_option': True})
        # Call it correctly and expect it doesn't raise anything
        self.class_being_tested(self.base_url, self.opener,
                                {'filename': 'hello', 'datafile_fields': '0',
                                 'datafile_format': 'irrelevant'})

    def upload_test(self):
        d = {'filename': self.file1,
             'datafile_fields': self.datafile_fields,
             'datafile_format': 'irrelevant'}
        d.update(self.datafiledict)
        df = self.class_being_tested(self.base_url, self.opener, d)
        df.update_database()
        read_timeseries(self.opener, self.base_url, self.ts1)
        read_timeseries(self.opener, self.base_url, self.ts2)
        self.assertEqual(len(self.ts1), 60)
        self.assertEqual(len(self.ts2), 60)
        (items1, items2, ritems1, ritems2) = [
            x.items() for x in (self.ts1, self.ts2,
                                self.ref_ts1, self.ref_ts2)]
        for i in range(0, 60):
            self.assertEqual(items1[i][0], ritems1[i][0])
            self.assertAlmostEqual(items1[i][1], ritems1[i][1], 4)
            self.assertEqual(items1[i][1].flags, ritems1[i][1].flags)
            self.assertEqual(items2[i][0], ritems2[i][0])
            self.assertAlmostEqual(items2[i][1], ritems2[i][1], 4)
            self.assertEqual(items2[i][1].flags, ritems2[i][1].flags)
        d['filename'] = full_testdata_filename(self.datafilename)
        df = self.class_being_tested(self.base_url, self.opener, d)
        df.update_database()
        read_timeseries(self.opener, self.base_url, self.ts1)
        read_timeseries(self.opener, self.base_url, self.ts2)
        self.assertEqual(len(self.ts1), 100)
        self.assertEqual(len(self.ts2), 100)
        (items1, items2, ritems1, ritems2) = [
            x.items() for x in (self.ts1, self.ts2,
                                self.ref_ts1, self.ref_ts2)]
        for i in range(0, 100):
            self.assertEqual(items1[i][0], ritems1[i][0])
            self.assertAlmostEqual(items1[i][1], ritems1[i][1], 4)
            self.assertEqual(items1[i][1].flags, ritems1[i][1].flags)
            self.assertEqual(items2[i][0], ritems2[i][0])
            self.assertAlmostEqual(items2[i][1], ritems2[i][1], 4)
            self.assertEqual(items2[i][1].flags, ritems2[i][1].flags)

    def runTest(self):
        if not self.class_being_tested or not self.base_url:
            return
        self.check_config_test()
        self.upload_test()


class TestDeltacom(_Test_logger):
    datafilename = 'deltacom_data.txt'
    class_being_tested = Datafile_deltacom


class TestSimple1(_Test_logger):
    datafilename = 'simple_data1.txt'
    class_being_tested = Datafile_simple
    datafiledict = {'date_format': '%y/%m/%d %H:%M:%S'}


class TestSimple2(_Test_logger):
    datafilename = 'simple_data2.txt'
    class_being_tested = Datafile_simple
    datafiledict = {'date_format': '%d/%m/%Y %H:%M:%S'}


class TestSimple3(_Test_logger):
    datafilename = 'simple_data3.txt'
    class_being_tested = Datafile_simple
    datafiledict = {'date_format': '%d/%m/%Y %H:%M',
                    'nfields_to_ignore': 1,
                    'delimiter': ','}


@skipUnless(os.getenv('PTHELMA_TEST_METEOLOGGER'),
            "set PTHELMA_TEST_METEOLOGGER")
class TestDst(TestCase):
    datafiledict = {'date_format': '%Y-%m-%dT%H:%M',
                    'timezone': 'Europe/Athens'}

    def setUp(self):
        get_server_from_env(self.__dict__)
        self.ref_ts = Timeseries(0)
        if not self.base_url:
            return
        self.opener = connect_to_server(self.base_url, self.username,
                                        self.password)
        self.timeseries_id = create_timeseries(self.opener, self.__dict__)
        self.ts = Timeseries(self.timeseries_id)

    def tearDown(self):
        if not self.base_url:
            return
        delete_timeseries(self.opener, self.base_url, self.timeseries_id)

    def run_test(self):
        if not self.base_url:
            return
        d = {'filename': full_testdata_filename(self.filename),
             'datafile_fields': str(self.timeseries_id),
             'datafile_format': 'irrelevant'}
        d.update(self.datafiledict)
        df = Datafile_simple(self.base_url, self.opener, d)
        df.update_database()
        read_timeseries(self.opener, self.base_url, self.ts)
        self.assertEqual(len(self.ts), len(self.ref_ts))
        (items, ritems) = [x.items() for x in (self.ts, self.ref_ts)]
        for item, ritem in zip(items, ritems):
            self.assertEqual(item[0], ritem[0])
            self.assertAlmostEqual(item[1], ritem[1], 4)
            self.assertEqual(item[1].flags, ritem[1].flags)

    @skipUnless(os.getenv('PTHELMA_TEST_METEOLOGGER'), "see above")
    def test_to_dst(self):
        self.filename = 'data_at_change_to_dst.txt'
        self.ref_ts.read(open(full_testdata_filename(
            'timeseries_at_change_to_dst.txt')))
        self.run_test()

    @skipUnless(os.getenv('PTHELMA_TEST_METEOLOGGER'), "see above")
    def test_from_dst(self):
        self.filename = 'data_at_change_from_dst.txt'
        self.ref_ts.read(open(full_testdata_filename(
            'timeseries_at_change_from_dst.txt')))
        self.run_test()


@skipUnless(os.getenv('PTHELMA_TEST_METEOLOGGER'),
            "set PTHELMA_TEST_METEOLOGGER")
class TestWdat5(_Test_logger):
    class_being_tested = Datafile_wdat5
    longMessage = True
    # The dicts below hold the parameter name as it is in the WeatherLink
    # README file, and the corresponding heading in the WeatherLink export file
    # (lowercase, after changing spaces to hyphens and removing dots), empty
    # string for those that aren't going to be tested.
    parameters = [
        {'name': 'outsideTemp', 'expname': 'temp-out'},
        {'name': 'hiOutsideTemp', 'expname': 'hi-temp'},
        {'name': 'lowOutsideTemp', 'expname': 'low-temp'},
        {'name': 'insideTemp', 'expname': 'in-temp'},
        {'name': 'barometer', 'expname': 'bar'},
        {'name': 'outsideHum', 'expname': 'out-hum'},
        {'name': 'insideHum', 'expname': 'in-hum'},
        {'name': 'rain', 'expname': 'rain'},
        {'name': 'hiRainRate', 'expname': 'rain-rate'},
        {'name': 'windSpeed', 'expname': 'wind-speed'},
        {'name': 'hiWindSpeed', 'expname': 'hi-speed'},
        {'name': 'windDirection', 'expname': 'wind-dir'},
        {'name': 'hiWindDirection', 'expname': 'hi-dir'},
        {'name': 'numWindSamples', 'expname': 'wind-samp'},
        {'name': 'solarRad', 'expname': 'solar-rad'},
        {'name': 'hiSolarRad', 'expname': 'hi-solar-rad'},
        {'name': 'UV', 'expname': 'uv-index'},
        {'name': 'hiUV', 'expname': 'hi-uv'},
        {'name': 'leafTemp1', 'expname': ''},
        {'name': 'leafTemp2', 'expname': ''},
        {'name': 'leafTemp3', 'expname': ''},
        {'name': 'leafTemp4', 'expname': ''},
        {'name': 'extraRad', 'expname': ''},
        {'name': 'newSensors1', 'expname': ''},
        {'name': 'newSensors2', 'expname': ''},
        {'name': 'newSensors3', 'expname': ''},
        {'name': 'newSensors4', 'expname': ''},
        {'name': 'newSensors5', 'expname': ''},
        {'name': 'newSensors6', 'expname': ''},
        {'name': 'forecast', 'expname': ''},
        {'name': 'ET', 'expname': 'et'},
        {'name': 'soilTemp1', 'expname': ''},
        {'name': 'soilTemp2', 'expname': ''},
        {'name': 'soilTemp3', 'expname': ''},
        {'name': 'soilTemp4', 'expname': ''},
        {'name': 'soilTemp5', 'expname': ''},
        {'name': 'soilTemp6', 'expname': ''},
        {'name': 'soilMoisture1', 'expname': ''},
        {'name': 'soilMoisture2', 'expname': ''},
        {'name': 'soilMoisture3', 'expname': ''},
        {'name': 'soilMoisture4', 'expname': ''},
        {'name': 'soilMoisture5', 'expname': ''},
        {'name': 'soilMoisture6', 'expname': ''},
        {'name': 'leafWetness1', 'expname': ''},
        {'name': 'leafWetness2', 'expname': ''},
        {'name': 'leafWetness3', 'expname': ''},
        {'name': 'leafWetness4', 'expname': ''},
        {'name': 'extraTemp1', 'expname': ''},
        {'name': 'extraTemp2', 'expname': ''},
        {'name': 'extraTemp3', 'expname': ''},
        {'name': 'extraTemp4', 'expname': ''},
        {'name': 'extraTemp5', 'expname': ''},
        {'name': 'extraTemp6', 'expname': ''},
        {'name': 'extraTemp7', 'expname': ''},
        {'name': 'extraHum1', 'expname': ''},
        {'name': 'extraHum2', 'expname': ''},
        {'name': 'extraHum3', 'expname': ''},
        {'name': 'extraHum4', 'expname': ''},
        {'name': 'extraHum5', 'expname': ''},
        {'name': 'extraHum6', 'expname': ''},
        {'name': 'extraHum7', 'expname': ''},
    ]

    def setUp(self):
        get_server_from_env(self.__dict__)

        # Connect to server
        self.opener = connect_to_server(self.base_url, self.username,
                                        self.password)

        # Create as many timeseries as there are not-to-ignore parameters
        for parm in self.parameters:
            parm['ts_id'] = create_timeseries(self.opener, self.__dict__
                                              ) if parm['expname'] else 0

    def tearDown(self):
        for parm in self.parameters:
            if parm['ts_id']:
                delete_timeseries(self.opener, self.base_url, parm['ts_id'])

    def check_config_test(self):
        self.assertRaises(ConfigurationError, self.class_being_tested,
                          self.base_url, self.opener, {})
        self.assertRaises(ConfigurationError, self.class_being_tested,
                          self.base_url, self.opener,
                          {'filename': 'hello', 'outsidetemp': '0',
                           'datafile_format': 'irrelevant',
                           'nonexistent_config_option': True})
        # Call it correctly and expect it doesn't raise anything
        self.class_being_tested(self.base_url, self.opener,
                                {'filename': 'hello', 'outsidetemp': '0',
                                 'datafile_format': 'irrelevant'})

    def upload_test(self):
        d = {'filename': full_testdata_filename('wdat5/1'),
             'datafile_format': 'irrelevant'}
        for parm in self.parameters:
            if parm['ts_id']:
                d[parm['name'].lower()] = parm['ts_id']
        df = Datafile_wdat5(self.base_url, self.opener, d)
        df.update_database()
        self.check(d['filename'])
        d['filename'] = full_testdata_filename('wdat5/2')
        df = Datafile_wdat5(self.base_url, self.opener, d)
        df.update_database()
        self.check(d['filename'])

    def check(self, datadir):
        for parm in self.parameters:
            if not parm['ts_id']:
                continue
            actual_ts = Timeseries(parm['ts_id'])
            read_timeseries(self.opener, self.base_url, actual_ts)
            reference_ts = Timeseries()
            with open(os.path.join(
                    datadir, 'generated', parm['expname'] + '.txt')) as f:
                reference_ts.read(f)
                precision = self.guess_precision(f)
            self.assertTimeseriesEqual(actual_ts, reference_ts, precision,
                                       parm['expname'] + '.txt')

    def guess_precision(self, f):
        result = 0
        f.seek(0)
        # Take a sample of 200 rows
        for i, line in enumerate(f):
            if i >= 200:
                break
            value = line.split(',')[1].strip()
            integral, sep, decimal = value.partition('.')
            result = len(decimal) if len(decimal) > result else result
        return result

    def assertTimeseriesEqual(self, ts1, ts2, precision, filename):
        self.assertEqual(len(ts1), len(ts2))
        items1 = ts1.items()
        items2 = ts2.items()
        for i, (item1, item2) in enumerate(zip(items1, items2)):
            self.assertEqual(item1[0], item2[0])
            if isnan(item1[1]):
                self.assertTrue(isnan(item2[1]),
                                msg='{1}, row {0}'.format(i + 1, filename))
            else:
                self.assertAlmostEqual(
                    item1[1], item2[1], places=precision,
                    msg='{1}, row {0}'.format(i + 1, filename))
            self.assertEqual(item1[1].flags, item2[1].flags)

    def assertAlmostEqual(self, a, b, places=7, msg=None):
        """We redefine assertAlmostEqual so that, e.g. with places=1, 0.55 is
        considered equal to both 0.5 and 0.6
        """
        tolerance = 10 ** (-places - 3)
        if abs(a - b) <= 0.5 * 10 ** (-places) + tolerance:
            return
        super(TestWdat5, self).assertAlmostEqual(a, b, places=places, msg=msg)



#!/usr/bin/python
"""
curves - curves interpolations

Copyright (C) 2005-2011 National Technical University of Athens
Copyright (C) 2011 Stefanos Kozanis

Based on openmeteo.thelma.source.interpol.pas writen in
Pascal by A. Christofides, Stefanos Kozanis

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

from math import log, exp
from datetime import datetime, MINYEAR, MAXYEAR
import math
from ConfigParser import ParsingError
from pthelma.timeseries import Timeseries, datetime_from_iso

class CurvePoint(object):
    def __init__(self, independent, dependent):
        self.independent = independent
        self.dependent = dependent

    def  __eq__(self, other):
        return abs(self.dependent-other.dependent)<0.001 and\
               abs(self.independent-other.independent)<0001

    def __str__(self):
        return '(%f,%f)'%(self.independent, self.dependent,)

    def __repr__(self):
        return '(%f,%f)'%(self.independent, self.dependent,)

    def v(self, reverse=False):
        if reverse:
            return (self.dependent, self.independent)
        return (self.independent, self.dependent)

errFUNCIONTACCEPTSCURVEPOINT = 'Function accept only CurvePoint '+\
                               'object as argument'

class InterpolatingCurve(object):
    def __init__(self, logarithmic=False, offset=CurvePoint(0,0)):
        self.logarithmic = logarithmic
        self.offset = offset
        self.container = []

    def add(self, *args):
        if len(args) not in (1,2):
            raise TypeError('add() takes one or two arguments '
                            '(%d given)'%(len(args),))  
        if len(args)==1:
            if not isinstance(args[0], CurvePoint):
                raise TypeError(errFUNCIONTACCEPTSCURVEPOINT)
            point = args[0]
        else:
            point = CurvePoint(*args)
        self.container.append(point)

    def __str__(self):
        return str(self.container)

    def __repr__(self):
        return str(self.container)

    def __getitem__(self, key):
        return self.container[key]

    def __setitem__(self, key, value):
        if not isinstance(value, CurvePoint):
            raise TypeError(errFUNCIONTACCEPTSCURVEPOINT)
        self.container[key] = value

    def __delitem__(self, key):
        del(self.container[key])

    def __len__(self):
        return len(self.container)

    def __getslice__(self, i, j):
        return self.container[i:j]

    def __setslice__(self, i, j, sequence):
        for item in sequence:
            if not isinstance(item, CurvePoint):
                raise TypeError(errFUNCIONTACCEPTSCURVEPOINT)
        self.container[i:j] = sequence

    def __delslice__(self, i, j):
        del(self.container[i:j])

    def index(self, value):
        if not isinstance(value, CurvePoint):
            raise TypeError(errFUNCIONTACCEPTSCURVEPOINT)
        return self.container.index(value)

    def insert(self, i, value):
        if not isinstance(value, CurvePoint):
            raise TypeError(errFUNCIONTACCEPTSCURVEPOINT)
        self.container[i:i] = [value]

    def first(self):
        return self.container[0]

    def last(self):
        return self.container[len(self.container)-1]

    def value_over_curve(self, value, reverse=False):
        return value>self.last().v(reverse)[0]

    def value_in_curve_range(self, value, reverse=False):
        return value<=self.last().v(reverse)[0] and\
               value>=self.first().v(reverse)[0]

    def read_line(self, line, columns=(0,1)):
        values = [float(x) for x in line.split(',')]
        self.add(*[values[i] for i in columns])

    def read_fp(self, fp, columns=(0,1)):
        line = fp.readline()
        while line:
            if line.isspace(): break
            self.read_line(line, columns)
            line = fp.readline()

    def interpolate(self, value, reverse=False):
        if self.last().v(reverse)[0]>self.first().v(reverse)[0]:
            i = len(self)-2
            while i>=0:
                if self[i].v(reverse)[0]<value:
                    break
                i-=1
            i = max(i, 0)
        else:
            i = len(self)-2
            while i>=0:
                if self[i].v(reverse)[0]>value:
                    break
                i-=1
            i = max(i, 0)
        if self.logarithmic and self[i] == self.offset:
            i+=1
        if self.logarithmic and self[i+1] == self.offset:
            i-=1
        if self.logarithmic:
            x0, y0 = self.offset.v(reverse)[0], self.offset.v(reverse)[1]
        else:
            x0, y0 = 0,0
        x1 = self[i].v(reverse)[0]+x0
        x2 = self[i+1].v(reverse)[0]+x0
        x  = value+x0
        y1 = self[i].v(reverse)[1]+y0
        y2 = self[i+1].v(reverse)[1]+y0
        if self.logarithmic:
            if x>0:
                return exp( (log(x)-log(x1))/(log(x2)-log(x1))*\
                            (log(y2)-log(y1)) + log(y1)) - y0
            else:
                return float('NaN')
        else:
            return (x-x1)/(x2-x1)*(y2-y1)+y1

    def reverseinterpolate(self, value):
        return self.interpolate(value, True)

MINDATE = datetime(MINYEAR, 1, 1, 0, 0, 0)
MAXDATE = datetime(MAXYEAR, 12, 31, 0, 0, 0)

def read_meta_line(fp):
    """Read one line from a file format header and return a (name, value)
    tuple, where name is lowercased. Returns ('', '') if the next line is
    blank. Raises ParsingError if next line in fp is not a valid header
    line."""
    line = fp.readline()
    (name, value) = '', ''
    if line.isspace(): return (name, value)
    if line.find('=') > 0:
        (name, value) = line.split('=', 1)
        name = name.rstrip().lower()
        value = value.strip()
    for c in name:
        if c.isspace():
            name = ''
            break
    if not name:
        raise ParsingError(("Invalid file header line"))
    return (name, value)

class TransientCurve(InterpolatingCurve):
    def __init__(self, logarithmic=False, offset=CurvePoint(0,0),
                 extension_line=False,
                 start_date=MINDATE, end_date=MAXDATE):
        super(TransientCurve, self).__init__(logarithmic, offset)
        self.start_date = start_date
        self.end_date = end_date
        self.extension_line=extension_line

    def read_fp(self, fp, columns=(0,1)):
        (name, value) = read_meta_line(fp)
        while name:
            if name=='extension':
                self.extension_line = str.lower(value)=='true'
            elif name=='logarithmic':
                self.logarithmic = str.lower(value)=='true'
            elif name=='offset':
                self.offset=CurvePoint(float(value),0)
            elif name in ('start_date', 'startdate'):
                self.start_date=datetime_from_iso(value)
            elif name in ('end_date', 'enddate'):
                self.end_date=datetime_from_iso(value)
            (name, value) = read_meta_line(fp)
        super(TransientCurve, self).read_fp(fp, columns)

class NoInterpolCurveError(Exception): pass

errFUNCIONTACCEPTSCURVE = 'Function accept only Transient Curve '+\
                          'object as argument'

class TransientCurveList(object):
    def __init__(self):
        self.container = []

    def addcurve(self, curve):
        if not isinstance(curve, TransientCurve):
            raise TypeError(errFUNCIONTACCEPTSCURVE)
        self.container.append(curve)

    def add(self, logarithmic=False, offset=CurvePoint(0,0),
                 extension_line=False,
                 start_date=MINDATE, end_date=MAXDATE):
        curve = TransientCurve(logarithmic, offset, extension_line,
                               start_date, end_date)
        self.addcurve(curve)

    def __str__(self):
        return str(self.container)

    def __repr__(self):
        return str(self.container)

    def __getitem__(self, key):
        return self.container[key]

    def __setitem__(self, key, value):
        if not isinstance(value, TransientCurve):
            raise TypeError(errFUNCIONTACCEPTSCURVE)
        self.container[key] = value

    def __delitem__(self, key):
        del(self.container[key])

    def __len__(self):
        return len(self.container)

    def __getslice__(self, i, j):
        return self.container[i:j]

    def __setslice__(self, i, j, sequence):
        for item in sequence:
            if not isinstance(item, TransientCurve):
                raise TypeError(errFUNCIONTACCEPTSCURVE)
        self.container[i:j] = sequence

    def __delslice__(self, i, j):
        del(self.container[i:j])

    def index(self, value):
        if not isinstance(value, TransientCurve):
            raise TypeError(errFUNCIONTACCEPTSCURVE)
        return self.container.index(value)

    def insert(self, i, value):
        if not isinstance(value, TransientCurve):
            raise TypeError(errFUNCIONTACCEPTSCURVE)
        self.container[i:i] = [value]

    def first(self):
        return self.container[0]

    def last(self):
        return self.container[len(self.container)-1]

    def find(self, date, extension_line=False):
        for curve in self.container:
            if date>=curve.start_date and date<=curve.end_date and\
               extension_line==curve.extension_line:
                return curve
        raise NoInterpolCurveError('No interpolation curve for '
                                   'date %s, with extension '
                                   'line=%s'%(date, extension_line,))

    def has_extension_lines(self):
        for curve in self.container:
            if curve.extension_line:
                return True
        return False

    def interpolate(self, date, value, reverse=False):
        if not self.has_extension_lines():
            return self.find(date).interpolate(value, reverse)
        normal_curve = self.find(date)
        if not normal_curve.value_over_curve(value, reverse):
            return normal_curve.interpolate(value, reverse)
        ext_curve = self.find(date, True)
        if ext_curve.value_in_curve_range(value, reverse) or \
                    ext_curve.value_over_curve(value, reverse):
            return ext_curve.interpolate(value, reverse)
        curve = InterpolatingCurve(logarithmic =\
                                            normal_curve.logarithmic,
                                   offset = normal_curve.offset)
        curve[0:2] = [normal_curve.last(), ext_curve.first()]
        return curve.interpolate(value, reverse)

    def interpolate_ts(self, timeseries, reverse=False):
        result = Timeseries(time_step=timeseries.time_step)
        for date in timeseries.iterkeys():
            value = timeseries[date]
            if math.isnan(value):
                result[date] = float('NaN')
            else:
                result[date] = self.interpolate(date, value, reverse)
        return result

    def stout_correction_series(self, stage, discharge):
        result = self.interpolate_ts(discharge, reverse=True)
        for date in result.iterkeys():
            value = timeseries[date]
            if (not math.isnan(value)) and (not math.isnan(stage[date])):
                result[date] = stage[date]-result[date]
            else:
                result[date] = float('NaN')
        return result

    def stout_correct_stage_series(self, timeseries, stage, discharge):

        def timedeltadivide(a, b):
            """Divide timedelta a by timedelta b."""
            a = a.days*86400+a.seconds
            b = b.days*86400+b.seconds
            return a/b

        correction_series = self.stout_correction_series(stage,
                                                         discharge)
        result = Timeseries(time_step=timeseries.time_step)
        assert(len(correction_series)>0)
        i1=i2=0
        d1=d2=correction_series.bounding_dates[0]
        v1=v2=correction_series[d1]
        for date in timeseries.iterkeys():
            if date>d1:
                i2+=1
                i1 = i2-1
                i2=max(i2, len(correction_series)-1)
                d1 = correction_series.items(pos=i1)[0]
                d2 = correction_series.items(pos=i2)[0]
                v1 = correction_series[d1]
                v2 = correction_series[d2]
            if d1!=d2:
                dh = v1*timedeltadivide(d2-date, d2-d1) +\
                     v2*timedeltadivide(date-d1, d2-d1)
            else:
                dh = v1
            if math.isnan(timeseries[date]):
                result[date]=float('NaN')
            else: 
                result[date]=timeseries[date]-dh
            if not self.has_extension_lines():
                continue
            normal_curve = self.find(date)
            if not normal_curve.value_over_curve(timeseries[date]):
                continue
            ext_curve = self.find(date, True)
            if ext_curve.value_in_curve_range(timeseries[date]) or \
                    ext_curve.value_over_curve(timeseries[date]):
                result[date]=timeseries[date]
                continue
            curve = InterpolatingCurve(logarithmic = False,
                                       offset = normal_curve.offset)
            curve[0:2] = [CurvePoint(normal_curve.last().v()[0],0), 
                          CurvePoint(ext_curve.first().v()[0],dh)]
            result[date]+=curve.interpolate(timeseries[date])

    def calc_stout_discharge(self, timeseries, stage, discharge):
        corrected_stage = stout_correct_stage_series(timeseries,
                                                     stage, discharge)
        return corrected_stage, interpolate_ts(corrected_stage)

    def read_fp(self, fp):
        (name, value) = read_meta_line(fp)
        while name:
            if name=='count': count = int(value)
            (name, value) = read_meta_line(fp)
        while count>0:
            self.add()
            self.last().read_fp(fp)
            count-=1




# copyright 2004 Michael D. Stenner <mstenner@ece.arizona.edu>
# license: LGPL

class xreverse:
    def __init__(self, file_object, buf_size=1024*8):
        self.fo = fo = file_object
        fo.seek(0, 2)        # go to the end of the file
        self.pos = fo.tell() # where we are 
        self.buffer = ''     # data buffer
        self.lbuf = []       # buffer for parsed lines
        self.done = 0        # we've read the last line
        self.jump = -1 * buf_size
        
        while 1:
            try:            fo.seek(self.jump, 1)
            except IOError: fo.seek(0)
            new_position = fo.tell()
            new = fo.read(self.pos - new_position)
            fo.seek(new_position)
            self.pos = new_position

            self.buffer = new + self.buffer
            if '\n' in new: break
            if self.pos == 0: return self.buffer

        nl = self.buffer.split('\n')
        nlb = [ i + '\n' for i in nl[1:-1] ]
        if not self.buffer[-1] == '\n': nlb.append(nl[-1])
        self.buffer = nl[0]
        self.lbuf = nlb

    def __iter__(self): return self

    def next(self):
        try:
            return self.lbuf.pop()
        except IndexError:
            fo = self.fo
            while 1:
                #get the next chunk of data
                try:            fo.seek(self.jump, 1)
                except IOError: fo.seek(0)
                new_position = fo.tell()
                new = fo.read(self.pos - new_position)
                fo.seek(new_position)
                self.pos = new_position

                nl = (new + self.buffer).split('\n')
                self.buffer = nl.pop(0)
                self.lbuf = [ i + '\n' for i in nl ]

                if self.lbuf: return self.lbuf.pop()
                elif self.pos == 0:
                    if self.done:
                        raise StopIteration
                    else:
                        self.done = 1
                        return self.buffer + '\n'

def dump(rtype, fn):
    import sys

    fo = file(fn)
    for line in rtype(fo):
        sys.stdout.write(line)
    fo.close()
    
def rereverse(rtype, fn):
    fo = file(fn)
    rev = [ line for line in xreverse(fo) ]
    rev.reverse()
    sys.stdout.writelines(rev)

def test_compare(rtype, fn):
    import sys, os

    t1 = os.times()

    fo = file(fn)
    for line in rtype(fo):
        pass
    fo.close()

    t2 = os.times()

    fo = file(fn)
    for line in fo.readlines():
        pass
    fo.close()

    t3 = os.times()

    for i in range(5):
        print t2[i] - t1[i], t3[i] - t2[i]
    


if __name__ == '__main__':
    import sys
    fn = sys.argv[1]
    #dump(xreverse, fn)
    #test_compare(xreverse, fn)
    rereverse(xreverse, fn)



#!/usr/bin/python
"""
tsprocess - time series processing

Copyright (C) 2005-2011 National Technical University of Athens
Copyright (C) 2011 Stefanos Kozanis, Antonis Christofides

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

from pthelma.timeseries import (Timeseries, 
                               timeseries_bounding_dates_from_db)
import math
from StringIO import StringIO
from pthelma.meteocalcs import HeatIndex, SSI, IDM, BarometricFormula
from pthelma.curves import (CurvePoint, TransientCurveList,
                            TransientCurve,)
from pthelma.datetimelist import DatetimeList
import copy


def GetTimeseriesCommonPeriod(tslist, start_date=None, end_date=None,
                              interval_exclusive=False, reject_nulls=True):
    assert(len(tslist)>0)
    common_period = DatetimeList()
    for date in tslist[0].iterkeys():
        if start_date and (date<start_date if not interval_exclusive\
                           else date<=start_date):
            continue
        if end_date and (date>end_date if not interval_exclusive\
                           else date>=end_date) :
            break
        value_rejected = False
        if reject_nulls and math.isnan(tslist[0][date]):
            continue
        for i in xrange(1, len(tslist)):
            if (not date in tslist[i]) or \
                       (reject_nulls and math.isnan(tslist[i][date])):
                value_rejected = True
                break
        if not value_rejected:
            common_period.append(date)
    return common_period


def MultiTimeseriesProcess(method, timeseries_arg, out_timeseries,
                           options={}):
    start_date = None
    end_date = None
    interval_exclusive = False
    if 'start_date' in options:
        start_date = options['start_date']
    if 'end_date' in options:
        end_date = options['end_date']
    if 'interval_exclusive' in options:
        interval_exclusive = options['interval_exclusive']
    common_period = GetTimeseriesCommonPeriod( timeseries_arg.values(),
                    start_date=start_date, end_date=end_date, 
                    interval_exclusive = interval_exclusive, reject_nulls=False)
    out_timeseries.clear()
    v = 0.0000
    first_timeseries = timeseries_arg[timeseries_arg.keys()[0]]
    for date in common_period:
        if method == 'HeatIndex':
            v = HeatIndex(timeseries_arg['temp'][date],
                          timeseries_arg['rh'][date])
        elif method == 'SSI':
            v = SSI(timeseries_arg['temp'][date],
                    timeseries_arg['rh'][date])
        elif method == 'IDM_monthly':
            v = IDM(timeseries_arg['temp'][date],
                    timeseries_arg['precip'][date], False)
        elif method == 'IDM_annual':
            v = IDM(timeseries_arg['temp'][date],
                    timeseries_arg['precip'][date], True)
        elif method == 'BaromFormula':
            v = BarometricFormula(timeseries_arg['temp'][date],
                    timeseries_arg['press'][date], options['hdiff'])
        elif method == 'OneStepDiff':
            prev_date = first_timeseries.time_step.previous(date)
            if prev_date<first_timeseries.bounding_dates()[0]:
                v = float('NaN')
            else:
                v = first_timeseries[date]-first_timeseries[prev_date]
        else:
            assert(False)
        out_timeseries[date] = v
        

def MultiTimeseriesProcessDb(method, timeseries_arg, out_timeseries_id,
                             db, read_tstep_func, transaction=None, 
                             commit=True, options={}):
    out_timeseries = Timeseries(id = out_timeseries_id)
    opts = copy.deepcopy(options)
    if 'append_only' in opts and opts['append_only']:
        bounds = timeseries_bounding_dates_from_db(db, 
                                                   id = out_timeseries_id)
        opts['start_date'] = bounds[1] if bounds else None;
        opts['interval_exclusive'] = True
    tseries_arg={}
    for key in timeseries_arg:
        ts = Timeseries(id=timeseries_arg[key])
        if ('append_only' in opts and opts['append_only']) \
                         and opts['start_date'] is not None:
            ts.read_from_db(db, bottom_only=True)
            if ts.bounding_dates()[0]>opts['start_date']:
                ts.read_from_db(db)
        else:
            ts.read_from_db(db)
        ts.time_step = read_tstep_func(ts.id)
        tseries_arg[key] = ts
    MultiTimeseriesProcess(method, tseries_arg, out_timeseries, opts)
    if 'append_only' in opts and opts['append_only']:
        out_timeseries.append_to_db(db=db, transaction=transaction,
                                    commit=commit)
    else:
        out_timeseries.write_to_db(db=db, transaction=transaction, 
                                   commit=commit)


def AggregateDbTimeseries(source_id, dest_id, db, read_tstep_func, transaction=None,
                          commit=True, missing_allowed=0.0,
                          missing_flag='MISSING', append_only=False,
                          last_incomplete=False, all_incomplete=False):
    source = Timeseries(id=source_id, time_step=read_tstep_func(source_id))
    dest_step = read_tstep_func(dest_id)
    if append_only:
        bounds = timeseries_bounding_dates_from_db(db = db, id = dest_id)
        end_date = bounds[1] if bounds else None
    source.read_from_db(db)
    dest = source.aggregate(target_step=dest_step,
                            missing_allowed=missing_allowed, 
                            missing_flag=missing_flag,
                            last_incomplete=last_incomplete,
                            all_incomplete=all_incomplete)[0]
    dest.id = dest_id
    if append_only:
        d=dest.bounding_dates()
        while (d is not None) and (end_date is not None) and d[0]<=end_date:
            del dest[d[0]]
            d=dest.bounding_dates()
        dest.append_to_db(db=db, transaction=transaction, commit=commit)
    else:
        dest.write_to_db(db=db, transaction=transaction, commit=commit)


def InterpolateDbTimeseries(source_id, dest_id, curve_type, curve_data,
                            db, data_columns=(0,1), logarithmic=False,
                            offset=0, append_only=False,
                            transaction=None, commit=True):
    if append_only:
        bounds = timeseries_bounding_dates_from_db(db, id = dest_id)
        start_date = bounds[1] if bounds else None;
    ts = Timeseries(id=source_id)
    if append_only and start_date is not None:
        ts.read_from_db(db, bottom_only=True)
        if ts.bounding_dates()[0]>start_date:
            ts.read_from_db(db)
        while ts.bounding_dates()[0]<=start_date:
            del(ts[ts.bounding_dates()[0]])
            if len(ts)==0: return
    else:
        ts.read_from_db(db)
    curve_list = TransientCurveList()
    if curve_type=='SingleCurve':
        curve_list.add(logarithmic=logarithmic, 
                       offset=CurvePoint(offset, 0))
        super(TransientCurve, 
              curve_list[0]).read_fp(StringIO(curve_data),
                                                data_columns)
    elif curve_type=='StageDischargeMulti':
        curve_list.read_fp(StringIO(curve_data))
    else:
        assert(False)
    out_timeseries = curve_list.interpolate_ts(ts)
    out_timeseries.id = dest_id
    if append_only:
        out_timeseries.append_to_db(db=db, transaction=transaction,
                                    commit=commit)
    else:
        out_timeseries.write_to_db(db=db, transaction=transaction, 
                                   commit=commit)





#!/usr/bin/env python
"""
loggertodb - insert automatic meteorological station data to the database

Copyright (C) 2005-2007 National Technical University of Athens
Copyright (C) 2005 Antonios Christofides

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

import sys
from datetime import datetime
from ConfigParser import RawConfigParser
import logging
from urllib2 import build_opener, HTTPCookieProcessor
import cookielib
import traceback

from pthelma import meteologger
from pthelma.meteologger import ConfigurationError


def check_configuration(config):
    required_options = ['base_url', 'user', 'password']
    optional_options = ['loglevel', 'logfile']
    all_options = required_options + optional_options
    for option in required_options:
        if not config.has_option('General', option):
            raise ConfigurationError('Option "{}" is required'.format(option))
    for option in config.options('General'):
        if option not in all_options:
            raise ConfigurationError('Unknown option "{}"'.format(option))
    warning_levels = ['ERROR', 'WARNING', 'INFO', 'DEBUG']
    if config.has_option('General', 'loglevel'):
        if config.get('General', 'loglevel') not in warning_levels:
            raise ConfigurationError('loglevel must be one of {}'.format(
                ', '.join(warning_levels)))


def read_configuration(config_file):
    defaults = (('General', 'loglevel', 'WARNING'),
                )
    config = RawConfigParser()
    config.read((config_file,))
    for d in defaults:
        if not config.has_option(d[0], d[1]):
            config.set(d[0], d[1], d[2])
    config.base_url = config.get('General', 'base_url')
    if not config.base_url.endswith('/'):
        config.base_url += '/'
    return config


def setup_logger(config):
    logger = logging.getLogger('loggertodb')
    logger.setLevel(logging.__dict__[config.get('General', 'loglevel')])
    if config.has_option('General', 'logfile'):
        logger.addHandler(logging.FileHandler(config.get('General',
                                                         'logfile')))
    else:
        logger.addHandler(logging.StreamHandler())
    return logger


def login_to_enhydris(config):
    cookiejar = cookielib.CookieJar()
    opener = build_opener(HTTPCookieProcessor(cookiejar))
    login_url = config.base_url + 'accounts/login/'
    opener.open(login_url)
    opener.addheaders = [('X-CSRFToken', cookie.value)
                         for cookie in cookiejar if cookie.name == 'csrftoken'
                         ] + [('Referer', login_url)]
    data = 'username={0}&password={1}'.format(
        config.get('General', 'user'), config.get('General', 'password'))
    opener.open(login_url, data)
    opener.addheaders = [('X-CSRFToken', cookie.value)
                         for cookie in cookiejar if cookie.name == 'csrftoken'
                         ]
    return opener


def execute():
    logger = None
    config = None
    try:
        if len(sys.argv) != 2:
            raise RuntimeError('usage: %s configfile' % (sys.argv[0]))
        config = read_configuration(sys.argv[1])
        logger = setup_logger(config)
        logger.info('Starting loggertodb, %s' % (datetime.today().isoformat()))
        opener = login_to_enhydris(config)
        sections = config.sections()[:]
        sections.remove('General')
        for section in sections:
            datafileclass = eval('meteologger.Datafile_%s' %
                                (config.get(section, 'datafile_format'),))
            adatafile = datafileclass(config.base_url, opener,
                                      dict(config.items(section)), logger)
            try:
                adatafile.update_database()
            except meteologger.MeteologgerError, e:
                msg = 'Error while processing item {0}: {1}'.format(section,
                                                                    str(e))
                sys.stderr.write(msg + '\n')
                logger.error(msg)
                logger.debug(traceback.format_exc())
    except Exception as e:
        msg = str(e)
        sys.stderr.write(msg + '\n')
        if logger:
            logger.error(msg)
        if config and config.get('General', 'loglevel') == 'DEBUG':
            raise
        sys.exit(1)
    finally:
        if logger:
            logger.info('Loggertodb finished, {}'.format(
                datetime.today().isoformat()))



from ConfigParser import RawConfigParser
from StringIO import StringIO
import textwrap
from unittest import TestCase

from pthelma.loggertodb import check_configuration
from pthelma.meteologger import ConfigurationError


class LoggertodbTestCase(TestCase):

    def test_check_configuration(self):
        cp = RawConfigParser()
        cp.readfp(StringIO(textwrap.dedent('''\
            [General]
            user = a_user
            password = a_password
            ''')))
        self.assertRaises(ConfigurationError, check_configuration, cp)

        cp = RawConfigParser()
        cp.readfp(StringIO(textwrap.dedent('''\
            [General]
            base_url = a_base_url
            user = a_user
            password = a_password
            nonexistent_option = an_option
            ''')))
        self.assertRaises(ConfigurationError, check_configuration, cp)

        cp = RawConfigParser()
        cp.readfp(StringIO(textwrap.dedent('''\
            [General]
            base_url = a_base_url
            user = a_user
            password = a_password
            loglevel = NONEXISTENT_LOG_LEVEL
            ''')))
        self.assertRaises(ConfigurationError, check_configuration, cp)

        # Call it correctly and expect it doesn't raise anything
        cp = RawConfigParser()
        cp.readfp(StringIO(textwrap.dedent('''\
            [General]
            base_url = a_base_url
            user = a_user
            password = a_password
            ''')))
        check_configuration(cp)

        # Again, with different options
        cp = RawConfigParser()
        cp.readfp(StringIO(textwrap.dedent('''\
            [General]
            base_url = a_base_url
            user = a_user
            password = a_password
            loglevel = ERROR
            ''')))
        check_configuration(cp)



#!/usr/bin/python
# -*- coding: utf8 -*-
"""
Tests for timeseries class.

NOTE: If you want to run the database tests, set the PSYCOPG_CONNECTION
      environment variable to "host=... dbname=... user=... password=...".


Copyright (C) 2005-2009 National Technical University of Athens
Copyright (C) 2005 Antonios Christofides

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

from datetime import datetime, timedelta
import math
import os
if os.getenv('PSYCOPG_CONNECTION'):
    import psycopg2
from StringIO import StringIO
import sys
import textwrap
from unittest import TestCase, skipUnless
import zlib

from pthelma.timeseries import TimeStep, strip_trailing_zeros, \
    datetime_from_iso, isoformat_nosecs, Timeseries, IntervalType, \
    identify_events


big_test_timeseries = textwrap.dedent("""\
            2003-07-18 18:53,93,\r
            2003-07-19 19:52,108.7,\r
            2003-07-20 23:59,28.3,HEARTS SPADES\r
            2003-07-21 00:02,,\r
            2003-07-22 00:02,,DIAMONDS\r
            2003-07-23 18:53,93,\r
            2003-07-24 19:52,108.7,\r
            2003-07-25 23:59,28.3,HEARTS SPADES\r
            2003-07-26 00:02,,\r
            2003-07-27 00:02,,DIAMONDS\r
            2003-08-18 18:53,93,\r
            2003-08-19 19:52,108.7,\r
            2003-08-20 23:59,28.3,HEARTS SPADES\r
            2003-08-21 00:02,,\r
            2003-08-22 00:02,,DIAMONDS\r
            2003-08-23 18:53,93,\r
            2003-08-24 19:52,108.7,\r
            2003-08-25 23:59,28.3,HEARTS SPADES\r
            2003-08-26 00:02,,\r
            2003-08-27 00:02,,DIAMONDS\r
            2004-07-18 18:53,93,\r
            2004-07-19 19:52,108.7,\r
            2004-07-20 23:59,28.3,HEARTS SPADES\r
            2004-07-21 00:02,,\r
            2004-07-22 00:02,,DIAMONDS\r
            2004-07-23 18:53,93,\r
            2004-07-24 19:52,108.7,\r
            2004-07-25 23:59,28.3,HEARTS SPADES\r
            2004-07-26 00:02,,\r
            2004-07-27 00:02,,DIAMONDS\r
            2004-08-18 18:53,93,\r
            2004-08-19 19:52,108.7,\r
            2004-08-20 23:59,28.3,HEARTS SPADES\r
            2004-08-21 00:02,,\r
            2004-08-22 00:02,,DIAMONDS\r
            2004-08-23 18:53,93,\r
            2004-08-24 19:52,108.7,\r
            2004-08-25 23:59,28.3,HEARTS SPADES\r
            2004-08-26 00:02,,\r
            2004-08-27 00:02,,DIAMONDS\r
            2005-07-18 18:53,93,\r
            2005-07-19 19:52,108.7,\r
            2005-07-20 23:59,28.3,HEARTS SPADES\r
            2005-07-21 00:02,,\r
            2005-07-22 00:02,,DIAMONDS\r
            2005-07-23 18:53,91,\r
            2005-07-24 19:52,1087,\r
            2005-07-25 23:59,-28.3,HEARTS SPADES\r
            2005-07-26 00:02,,\r
            2005-07-27 00:02,,DIAMONDS\r
            2005-08-18 18:53,95,\r
            2005-08-19 19:52,110.7,\r
            2005-08-20 23:59,8.3,HEARTS SPADES\r
            2005-08-21 00:02,,\r
            2005-08-22 00:02,,DIAMONDS\r
            2005-08-23 18:53,94,\r
            2005-08-24 19:52,109.7,\r
            2005-08-25 23:59,27.3,HEARTS SPADES\r
            2005-08-26 00:02,,\r
            2005-08-27 00:02,,DIAMONDS\r
            """)


tenmin_test_timeseries = textwrap.dedent("""\
            2008-02-07 09:40,10.32,
            2008-02-07 09:50,10.42,
            2008-02-07 10:00,10.51,
            2008-02-07 10:10,10.54,
            2008-02-07 10:20,10.71,
            2008-02-07 10:30,10.96,
            2008-02-07 10:40,10.93,
            2008-02-07 10:50,11.10,
            2008-02-07 11:00,11.23,
            2008-02-07 11:10,11.44,
            2008-02-07 11:20,11.41,
            2008-02-07 11:30,11.42,MISS
            2008-02-07 11:40,11.54,
            2008-02-07 11:50,11.68,
            2008-02-07 12:00,11.80,
            2008-02-07 12:10,11.91,
            2008-02-07 12:20,12.16,
            2008-02-07 12:30,12.16,
            2008-02-07 12:40,12.24,
            2008-02-07 12:50,12.13,
            2008-02-07 13:00,12.17,
            2008-02-07 13:10,12.31,
            """)


tenmin_allmiss_test_timeseries = textwrap.dedent("""\
            2005-05-01 00:10,,
            2005-05-01 00:20,,
            2005-05-01 00:30,,
            2005-05-01 00:40,,
            2005-05-01 00:50,,
            2005-05-01 01:00,,
            2005-05-01 01:10,,
            2005-05-01 01:20,1,
            2005-05-01 01:30,,
            2005-05-01 01:40,1,
            2005-05-01 01:50,,
            2005-05-01 02:00,1,
            """)


tenmin_vector_test_timeseries = textwrap.dedent("""\
            2005-05-01 00:10,350,
            2005-05-01 00:20,355,
            2005-05-01 00:30,358,
            2005-05-01 00:40,2,
            2005-05-01 00:50,10,
            2005-05-01 01:00,5,
            2005-05-01 01:10,272,
            2005-05-01 01:20,268,
            2005-05-01 01:30,275,
            2005-05-01 01:40,265,
            2005-05-01 01:50,264,
            2005-05-01 02:00,276,
            """)


tenmin_test_timeseries_file = textwrap.dedent(u"""\
            Version=2\r
            Unit=°C\r
            Count=22\r
            Title=A test 10-min time series\r
            Comment=This timeseries is extremely important\r
            Comment=because the comment that describes it\r
            Comment=spans five lines.\r
            Comment=\r
            Comment=These five lines form two paragraphs.\r
            Timezone=EET (UTC+0200)\r
            Time_step=10,0\r
            Nominal_offset=0,0\r
            Actual_offset=0,0\r
            Variable=temperature\r
            Precision=1\r
            \r
            2008-02-07 09:40,10.3,\r
            2008-02-07 09:50,10.4,\r
            2008-02-07 10:00,10.5,\r
            2008-02-07 10:10,10.5,\r
            2008-02-07 10:20,10.7,\r
            2008-02-07 10:30,11.0,\r
            2008-02-07 10:40,10.9,\r
            2008-02-07 10:50,11.1,\r
            2008-02-07 11:00,11.2,\r
            2008-02-07 11:10,11.4,\r
            2008-02-07 11:20,11.4,\r
            2008-02-07 11:30,11.4,MISS\r
            2008-02-07 11:40,11.5,\r
            2008-02-07 11:50,11.7,\r
            2008-02-07 12:00,11.8,\r
            2008-02-07 12:10,11.9,\r
            2008-02-07 12:20,12.2,\r
            2008-02-07 12:30,12.2,\r
            2008-02-07 12:40,12.2,\r
            2008-02-07 12:50,12.1,\r
            2008-02-07 13:00,12.2,\r
            2008-02-07 13:10,12.3,\r
            """)


aggregated_hourly_sum = textwrap.dedent("""\
            2008-02-07 10:00,31.25,MISS\r
            2008-02-07 11:00,65.47,\r
            2008-02-07 12:00,69.29,MISS\r
            2008-02-07 13:00,72.77,\r
            """)


aggregated_hourly_average = textwrap.dedent("""\
            2008-02-07 10:00,10.4166666666667,MISS\r
            2008-02-07 11:00,10.911666667,\r
            2008-02-07 12:00,11.548333333,MISS\r
            2008-02-07 13:00,12.128333333,\r
            """)


aggregated_hourly_max = textwrap.dedent("""\
            2008-02-07 10:00,10.51,MISS\r
            2008-02-07 11:00,11.23,\r
            2008-02-07 12:00,11.8,MISS\r
            2008-02-07 13:00,12.24,\r
            """)


aggregated_hourly_min = textwrap.dedent("""\
            2008-02-07 10:00,10.32,MISS\r
            2008-02-07 11:00,10.54,\r
            2008-02-07 12:00,11.41,MISS\r
            2008-02-07 13:00,11.91,\r
            """)


aggregated_hourly_missing = textwrap.dedent("""\
            2008-02-07 10:00,3,\r
            2008-02-07 11:00,0,\r
            2008-02-07 12:00,0,\r
            2008-02-07 13:00,0,\r
            """)


aggregated_hourly_vector_average = textwrap.dedent("""\
            2005-05-01 01:00,0,\r
            2005-05-01 02:00,270,\r
            """)


aggregated_hourly_allmiss = textwrap.dedent("""\
            2005-05-01 02:00,3,MISS\r
            """)


aggregated_hourly_vector_missing = textwrap.dedent("""\
            2005-05-01 01:00,0,\r
            2005-05-01 02:00,0,\r
            """)


def extract_lines(s, from_, to):
    return ''.join(s.splitlines(True)[from_:to])


class _Test_timestep_utilities(TestCase):

    def setUp(self):
        self.dailystep = TimeStep(length_minutes=1440,
                                  nominal_offset=(480, 0),
                                  actual_offset=(0, 0))
        self.bimonthlystep = TimeStep(length_months=2,
                                      nominal_offset=(0, 0),
                                      actual_offset=(-1442, 2))

    def test_up1(self):
        self.assertEqual(self.dailystep.up(datetime(1964, 2, 29, 18, 35)),
                         datetime(1964, 3, 1, 8, 0))

    def test_up2(self):
        self.assertEqual(self.dailystep.up(datetime(1964, 3, 1, 7, 59)),
                         datetime(1964, 3, 1, 8, 0))

    def test_up3(self):
        self.assertEqual(self.dailystep.up(datetime(1964, 3, 1, 8, 0)),
                         datetime(1964, 3, 1, 8, 0))

    def test_up4(self):
        self.assertEqual(self.bimonthlystep.up(datetime(1964, 2, 29, 18, 35)),
                         datetime(1964, 3, 1, 0, 0))

    def test_up5(self):
        self.assertEqual(self.bimonthlystep.up(datetime(1963, 12, 31, 18, 35)),
                         datetime(1964, 1, 1, 0, 0))

    def test_down1(self):
        self.assertEqual(self.dailystep.down(datetime(1964, 2, 29, 18, 35)),
                         datetime(1964, 2, 29, 8, 0))

    def test_down2(self):
        self.assertEqual(self.dailystep.down(datetime(1964, 3, 1, 7, 59)),
                         datetime(1964, 2, 29, 8, 0))

    def test_down3(self):
        self.assertEqual(self.dailystep.down(datetime(1964, 3, 1, 8, 0)),
                         datetime(1964, 3, 1, 8, 0))

    def test_down4(self):
        self.assertEqual(self.bimonthlystep.down(
            datetime(1964, 2, 29, 18, 35)), datetime(1964, 1, 1, 0, 0))

    def test_down5(self):
        self.assertEqual(self.bimonthlystep.down(
            datetime(1963, 12, 31, 18, 35)), datetime(1963, 11, 1, 0, 0))

    def test_next1(self):
        self.assertEqual(self.dailystep.next(datetime(1964, 2, 29, 15, 12)),
                         datetime(1964, 3, 2, 8, 0))

    def test_next2(self):
        self.assertEqual(self.dailystep.next(datetime(1964, 3, 2, 8, 0)),
                         datetime(1964, 3, 3, 8, 0))

    def test_next3(self):
        self.assertEqual(self.bimonthlystep.next(
            datetime(1964, 2, 29, 15, 12)), datetime(1964, 5, 1, 0, 0))

    def test_next4(self):
        self.assertEqual(self.bimonthlystep.next(datetime(1964, 5, 1, 0, 0)),
                         datetime(1964, 7, 1, 0, 0))

    def test_previous1(self):
        self.assertEqual(self.dailystep.previous(
            datetime(1964, 2, 29, 15, 12)), datetime(1964, 2, 28, 8, 0))

    def test_previous2(self):
        self.assertEqual(self.dailystep.previous(datetime(1964, 2, 29, 8, 0)),
                         datetime(1964, 2, 28, 8, 0))

    def test_previous3(self):
        self.assertEqual(self.bimonthlystep.previous(
            datetime(1964, 2, 29, 15, 12)), datetime(1963, 11, 1, 0, 0))

    def test_previous4(self):
        self.assertEqual(self.bimonthlystep.previous(
            datetime(1963, 11, 1, 0, 0)), datetime(1963, 9, 1, 0, 0))

    def test_actual_timestamp1(self):
        self.assertEqual(self.dailystep.actual_timestamp(
            datetime(1964, 2, 29, 8, 0)), datetime(1964, 2, 29, 8, 0))

    def test_actual_timestamp2(self):
        self.assertEqual(self.bimonthlystep.actual_timestamp(
            datetime(1964, 3, 1, 0, 0)), datetime(1964, 4, 29, 23, 58))

    def test_containing_interval1(self):
        self.assertEqual(self.dailystep.containing_interval(datetime(
            1964, 2, 29, 15, 12)), datetime(1964, 3, 1, 8, 0))

    def test_containing_interval2(self):
        self.assertEqual(self.bimonthlystep.containing_interval(datetime(
            1964, 2, 29, 15, 12)), datetime(1964, 3, 1, 0, 0))

    def test_containing_interval3(self):
        self.assertEqual(self.bimonthlystep.containing_interval(datetime(
            1964, 4, 18, 15, 12)), datetime(1964, 3, 1, 0, 0))

    def test_interval_endpoints1(self):
        self.assertEqual(self.bimonthlystep.interval_endpoints(datetime(
            1964, 3, 1, 0, 0)),
            (datetime(1964, 2, 28, 23, 58), datetime(1964, 4, 29, 23, 58)))


class _Test_datetime_utilities(TestCase):

    def setUp(self):
        self.d = datetime(1964, 2, 29, 18, 35)

    def test_datetime_from_iso1(self):
        self.assertEqual(self.d, datetime_from_iso("1964-02-29 18:35"))

    def test_datetime_from_iso2(self):
        self.assertEqual(self.d, datetime_from_iso("1964-02-29T18:35"))

    def test_datetime_from_iso3(self):
        self.assertEqual(self.d, datetime_from_iso("1964-02-29t18:35"))

    def test_datetime_from_iso4(self):
        self.assertRaises(ValueError, datetime_from_iso, "1964")

    def test_datetime_from_iso5(self):
        self.assertRaises(ValueError, datetime_from_iso, "1965-29-29 18:35")

    def test_date_time_from_iso6(self):
        self.assertEqual(datetime(1964, 2, 29),
                         datetime_from_iso("1964-02-29"))

    def test_isoformat_nosecs1(self):
        self.assertEqual(isoformat_nosecs(self.d), "1964-02-29T18:35")

    def test_isoformat_nosecs2(self):
        self.assertEqual(isoformat_nosecs(self.d, ' '), "1964-02-29 18:35")


class _Test_strip_trailing_zeros(TestCase):

    def test1(self):
        self.assertEqual(strip_trailing_zeros('hello'), 'hello')

    def test2(self):
        self.assertEqual(strip_trailing_zeros('hello0'), 'hello0')

    def test3(self):
        self.assertEqual(strip_trailing_zeros('850'), '850')

    def test4(self):
        self.assertEqual(strip_trailing_zeros('8500'), '8500')

    def test5(self):
        self.assertEqual(strip_trailing_zeros('0.0'), '0')

    def test6(self):
        self.assertEqual(strip_trailing_zeros('0.00'), '0')

    def test7(self):
        self.assertEqual(strip_trailing_zeros('18.3500'), '18.35')

    def test8(self):
        self.assertEqual(strip_trailing_zeros('18.000'), '18')


class _Test_Timeseries_setitem(TestCase):

    def setUp(self):
        self.ts = Timeseries()
        self.date = datetime_from_iso("1970-05-23 08:00")

    def test1(self):
        self.ts[self.date] = 8.2
        self.assertEqual(self.ts[self.date], 8.2)
        self.assertEqual(self.ts[self.date].flags, set([]))

    def test2(self):
        self.ts[self.date] = (9.3, ['RANGE', 'SPADE'])
        self.assertEqual(self.ts[self.date], 9.3)
        self.assertEqual(self.ts[self.date].flags, set(['RANGE', 'SPADE']))
        self.ts[self.date] = 8.2
        self.assertEqual(self.ts[self.date], 8.2)
        self.assertEqual(self.ts[self.date].flags, set(['RANGE', 'SPADE']))

    def test3(self):
        self.ts[self.date] = float('NaN')
        self.assert_(math.isnan(self.ts[self.date]))


class _Test_Timeseries_write_empty(TestCase):

    def setUp(self):
        self.ts = Timeseries()
        self.c = StringIO()

    def test(self):
        self.ts.write(self.c)
        self.assertEqual(self.c.getvalue(), '')


class _Test_Timeseries_write_nonempty(TestCase):

    def setUp(self):
        self.ts = Timeseries()
        self.c = StringIO()
        self.ts["2005-08-23 18:53"] = 93
        self.ts["2005-08-24 19:52"] = 108.7
        self.ts["2005-08-25 23:59"] = (28.3, ['HEARTS', 'SPADES'])
        self.ts["2005-08-26 00:02"] = float('NaN')
        self.ts["2005-08-27 00:02"] = (float('NaN'), ['DIAMONDS'])

    def test_all(self):
        self.ts.write(self.c)
        self.assertEqual(self.c.getvalue(), textwrap.dedent("""\
            2005-08-23 18:53,93,\r
            2005-08-24 19:52,108.7,\r
            2005-08-25 23:59,28.3,HEARTS SPADES\r
            2005-08-26 00:02,,\r
            2005-08-27 00:02,,DIAMONDS\r
            """))

    def test_with_start1(self):
        self.ts.write(self.c, start=datetime_from_iso("2005-08-24 19:52"))
        self.assertEqual(self.c.getvalue(), textwrap.dedent("""\
            2005-08-24 19:52,108.7,\r
            2005-08-25 23:59,28.3,HEARTS SPADES\r
            2005-08-26 00:02,,\r
            2005-08-27 00:02,,DIAMONDS\r
            """))

    def test_with_start2(self):
        self.ts.write(self.c, start=datetime_from_iso("2005-08-24 19:51"))
        self.assertEqual(self.c.getvalue(), textwrap.dedent("""\
            2005-08-24 19:52,108.7,\r
            2005-08-25 23:59,28.3,HEARTS SPADES\r
            2005-08-26 00:02,,\r
            2005-08-27 00:02,,DIAMONDS\r
            """))

    def test_with_start3(self):
        self.ts.write(self.c, start=datetime_from_iso("2000-08-24 19:51"))
        self.assertEqual(self.c.getvalue(), textwrap.dedent("""\
            2005-08-23 18:53,93,\r
            2005-08-24 19:52,108.7,\r
            2005-08-25 23:59,28.3,HEARTS SPADES\r
            2005-08-26 00:02,,\r
            2005-08-27 00:02,,DIAMONDS\r
            """))

    def test_with_end1(self):
        self.ts.write(self.c, end=datetime_from_iso("2005-08-27 00:02"))
        self.assertEqual(self.c.getvalue(), textwrap.dedent("""\
            2005-08-23 18:53,93,\r
            2005-08-24 19:52,108.7,\r
            2005-08-25 23:59,28.3,HEARTS SPADES\r
            2005-08-26 00:02,,\r
            2005-08-27 00:02,,DIAMONDS\r
            """))

    def test_with_end2(self):
        self.ts.write(self.c, end=datetime_from_iso("3005-08-27 00:02"))
        self.assertEqual(self.c.getvalue(), textwrap.dedent("""\
            2005-08-23 18:53,93,\r
            2005-08-24 19:52,108.7,\r
            2005-08-25 23:59,28.3,HEARTS SPADES\r
            2005-08-26 00:02,,\r
            2005-08-27 00:02,,DIAMONDS\r
            """))

    def test_with_end3(self):
        self.ts.write(self.c, end=datetime_from_iso("2005-08-26 00:03"))
        self.assertEqual(self.c.getvalue(), textwrap.dedent("""\
            2005-08-23 18:53,93,\r
            2005-08-24 19:52,108.7,\r
            2005-08-25 23:59,28.3,HEARTS SPADES\r
            2005-08-26 00:02,,\r
            """))

    def test_with_start_and_end1(self):
        self.ts.write(self.c, start=datetime_from_iso("2005-08-24 19:50"),
                      end=datetime_from_iso("2005-08-26 00:03"))
        self.assertEqual(self.c.getvalue(), textwrap.dedent("""\
            2005-08-24 19:52,108.7,\r
            2005-08-25 23:59,28.3,HEARTS SPADES\r
            2005-08-26 00:02,,\r
            """))

    def test_with_start_and_end2(self):
        self.ts.write(self.c, start=datetime_from_iso("2005-08-24 19:53"),
                      end=datetime_from_iso("2005-08-24 19:54"))
        self.assertEqual(self.c.getvalue(), "")


class _Test_Timeseries_file(TestCase):

    def setUp(self):
        self.reference_ts = Timeseries(
            time_step=TimeStep(length_minutes=10, length_months=0,
                               nominal_offset=(0, 0)),
            unit=u'°C', title=u"A test 10-min time series", precision=1,
            timezone=u"EET (UTC+0200)", variable=u"temperature",
            comment="This timeseries is extremely important\n"
                    "because the comment that describes it\n"
                    "spans five lines.\n\n"
                    "These five lines form two paragraphs.")
        self.reference_ts.read(StringIO(tenmin_test_timeseries))

    def test_write_file(self):
        outstring = StringIO()
        self.reference_ts.write_file(outstring)
        self.assertEqual(outstring.getvalue(), tenmin_test_timeseries_file)

    def test_read_file(self):
        instring = StringIO(tenmin_test_timeseries_file)
        ts = Timeseries()
        ts.read_file(instring)
        self.assertEqual(ts.time_step.length_minutes,
                         self.reference_ts.time_step.length_minutes)
        self.assertEqual(ts.time_step.length_months,
                         self.reference_ts.time_step.length_months)
        self.assertEqual(ts.time_step.nominal_offset,
                         self.reference_ts.time_step.nominal_offset)
        self.assertEqual(ts.time_step.actual_offset,
                         self.reference_ts.time_step.actual_offset)
        self.assertEqual(ts.time_step.interval_type,
                         self.reference_ts.time_step.interval_type)
        self.assertEqual(ts.unit, self.reference_ts.unit)
        self.assertEqual(ts.title, self.reference_ts.title)
        self.assertEqual(ts.precision, self.reference_ts.precision)
        self.assertEqual(ts.timezone, self.reference_ts.timezone)
        self.assertEqual(ts.variable, self.reference_ts.variable)
        self.assertEqual(ts.comment, self.reference_ts.comment)
        self.assertEqual(len(ts), len(self.reference_ts))
        for d in self.reference_ts:
            self.assertAlmostEqual(ts[d], self.reference_ts[d], 1)


class _Test_Timeseries_read(TestCase):

    # This test trusts Timeseries.write, so it must be independently tested.
    def check(self, s):
        ts = Timeseries()
        instring = StringIO(s)
        outstring = StringIO()
        ts.read(instring)
        ts.write(outstring)
        self.assertEqual(outstring.getvalue(), instring.getvalue())

    def test_read_empty(self):
        self.check('')

    def test_read_nonempty(self):
        self.check(textwrap.dedent("""\
            2005-08-23 18:53,93,\r
            2005-08-24 19:52,108.7,\r
            2005-08-25 23:59,28.3,HEARTS SPADES\r
            2005-08-26 00:02,,\r
            2005-08-27 00:02,,DIAMONDS\r
            """))

    def test_read_emptylines(self):
        instring = StringIO(textwrap.dedent("""\
            \r
            \r
            2005-08-23 18:53,93,\r
            \r
            \r
            2005-08-24 19:52,108.7,\r
            \r
            \r
            """))
        cleanstring = textwrap.dedent("""\
            2005-08-23 18:53,93,\r
            2005-08-24 19:52,108.7,\r
            """)
        outstring = StringIO()
        ts = Timeseries()
        ts.read(instring)
        ts.write(outstring)
        self.assertEqual(outstring.getvalue(), cleanstring)

    def test_error1(self):
        self.assertRaises(ValueError, Timeseries.read, Timeseries(),
                          StringIO('85'))

    def test_error2(self):
        self.assertRaises(ValueError, Timeseries.read, Timeseries(),
                          StringIO("""\
            2005-08-23 18:53,93,
            2005-08-24 19:52,108.7,
            2005-08-25 23:59,28.3,HEARTS SPADES
            2005-08-26 00:02,
            2005-08-27 00:02,,DIAMONDS
            """))  # Has only one comma in line four

    def test_error3(self):
        self.assertRaises(ValueError, Timeseries.read, Timeseries(),
                          StringIO('abcd,,'))


@skipUnless(os.getenv("PSYCOPG_CONNECTION"), "set PSYCOPG_CONNECTION")
class _Test_Timeseries_write_to_db(TestCase):

    def setUp(self):
        self.db = psycopg2.connect(os.getenv("PSYCOPG_CONNECTION"))
        c = self.db.cursor()
        c.execute('SET CONSTRAINTS ALL DEFERRED')
        c.close()

    def tearDown(self):
        self.db.rollback()
        self.db.close()

    def test_bottom_only(self):
        ts = Timeseries(0)
        ts.read(StringIO(textwrap.dedent("""\
            2005-08-23 18:53,93,
            2005-08-24 19:52,108.7,
            2005-08-25 23:59,28.3,HEARTS SPADES
            2005-08-26 00:02,,
            2005-08-27 00:02,,DIAMONDS
            """)))
        ts.write_to_db(self.db, commit=False)
        c = self.db.cursor()
        c.execute("""SELECT top, middle, bottom FROM ts_records
                     WHERE id=%d""" % (ts.id))
        (top, middle, bottom) = c.fetchone()
        c.close()
        sys.stderr.write('------------------\n' + top
                         + '--------------------\n')
        self.assertFalse(top)
        self.assert_(middle is None)
        self.assertEqual(bottom, textwrap.dedent("""\
            2005-08-23 18:53,93,\r
            2005-08-24 19:52,108.7,\r
            2005-08-25 23:59,28.3,HEARTS SPADES\r
            2005-08-26 00:02,,\r
            2005-08-27 00:02,,DIAMONDS\r
            """))

    def test_top_middle_bottom(self):
        saved_max_all_bottom = Timeseries.MAX_BOTTOM
        Timeseries.MAX_ALL_BOTTOM = 20
        try:
            ts = Timeseries(0)
            ts.read(StringIO(big_test_timeseries))
            ts.write_to_db(self.db, commit=False)
            c = self.db.cursor()
            c.execute("""SELECT top, middle, bottom FROM ts_records
                         WHERE id=%d""" % (ts.id))
            (db_top, db_middle, db_bottom) = c.fetchone()
            c.close()
            actual_top = extract_lines(big_test_timeseries, None,
                                       Timeseries.ROWS_IN_TOP_BOTTOM)
            actual_bottom = extract_lines(big_test_timeseries,
                                          -Timeseries.ROWS_IN_TOP_BOTTOM, None)
            actual_middle = extract_lines(big_test_timeseries,
                                          Timeseries.ROWS_IN_TOP_BOTTOM,
                                          -Timeseries.ROWS_IN_TOP_BOTTOM)
            self.assertEqual(db_top, actual_top)
            self.assertEqual(zlib.decompress(str(db_middle)), actual_middle)
            self.assertEqual(db_bottom, actual_bottom)
        finally:
            Timeseries.MAX_ALL_BOTTOM = saved_max_all_bottom


@skipUnless(os.getenv("PSYCOPG_CONNECTION"), "set PSYCOPG_CONNECTION")
class _Test_Timeseries_read_from_db(TestCase):
    # This test trusts Timeseries.write_to_db, so it must be independently
    # tested

    def setUp(self):
        self.db = psycopg2.connect(os.getenv("PSYCOPG_CONNECTION"))
        c = self.db.cursor()
        c.execute('SET CONSTRAINTS ALL DEFERRED')
        c.close()

    def tearDown(self):
        self.db.rollback()
        self.db.close()

    def check(self, s):
        ts1 = Timeseries(0)
        instring = StringIO(s)
        outstring = StringIO()
        ts1.read(instring)
        ts1.write_to_db(self.db, commit=False)
        ts2 = Timeseries(0)
        ts2.read_from_db(self.db)
        ts2.write(outstring)
        self.assertEqual(outstring.getvalue(), instring.getvalue())

    def test_read_empty(self):
        self.check('')

    def test_read_nonempty_small(self):
        self.check(textwrap.dedent("""\
            2005-08-23 18:53,93,\r
            2005-08-24 19:52,108.7,\r
            2005-08-25 23:59,28.3,HEARTS SPADES\r
            2005-08-26 00:02,,\r
            2005-08-27 00:02,,DIAMONDS\r
            """))

    def test_read_nonempty_big(self):
        self.check(big_test_timeseries)


class _Test_Timeseries_append(TestCase):

    def test_append(self):
        ts1 = Timeseries()
        ts1.read(StringIO(extract_lines(big_test_timeseries, 0, 5)))
        ts2 = Timeseries()
        ts2.read(StringIO(extract_lines(big_test_timeseries, 5, 10)))
        ts1.append(ts2)
        out = StringIO()
        ts1.write(out)
        self.assertEqual(out.getvalue(),
                         extract_lines(big_test_timeseries, 0, 10))
        ts3 = Timeseries()
        ts3.read(StringIO(extract_lines(big_test_timeseries, 9, 10)))
        self.assertRaises(Exception, Timeseries.append, ts1, ts3)


@skipUnless(os.getenv("PSYCOPG_CONNECTION"), "set PSYCOPG_CONNECTION")
class _Test_Timeseries_append_to_db(TestCase):
    # This trusts everything else, so no other test should trust append and
    # append_to_db.

    def setUp(self):
        self.db = psycopg2.connect(os.getenv("PSYCOPG_CONNECTION"))
        c = self.db.cursor()
        c.execute('SET CONSTRAINTS ALL DEFERRED')
        c.close()

    def tearDown(self):
        self.db.rollback()
        self.db.close()

    def test_append_to_db(self):
        out = StringIO()
        ts1 = Timeseries(0)
        ts1.read(StringIO(extract_lines(big_test_timeseries, 0, 5)))
        ts1.write_to_db(self.db, commit=False)
        ts1.clear()
        ts1.read(StringIO(extract_lines(big_test_timeseries, 5, 10)))
        ts1.append_to_db(self.db, commit=False)
        ts1.clear()
        ts1.read_from_db(self.db)
        ts1.write(out)
        self.assertEqual(out.getvalue(),
                         extract_lines(big_test_timeseries, 0, 10))
        ts1.clear()
        ts1.read(StringIO(extract_lines(big_test_timeseries, 10, None)))
        ts1.append_to_db(self.db, commit=False)
        ts1.clear()
        ts1.read_from_db(self.db)
        out.truncate(0)
        ts1.write(out)
        self.assertEqual(out.getvalue(), big_test_timeseries)
        ts1.clear()
        ts1.read(StringIO(extract_lines(big_test_timeseries, 10, None)))
        self.assertRaises(Exception, Timeseries.append_to_db, ts1, self.db,
                          commit=False)


class _Test_Timeseries_item(TestCase):

    def setUp(self):
        self.ts = Timeseries()
        self.ts.read(StringIO(big_test_timeseries))

    def test_matches_exactly(self):
        item = self.ts.item('2003-07-24 19:52')
        self.assertAlmostEqual(item[1], 108.7)
        self.assertEqual(isoformat_nosecs(item[0]), '2003-07-24T19:52')

    def test_matches_next(self):
        item = self.ts.item('2003-07-24 00:00')
        self.assertAlmostEqual(item[1], 108.7)
        self.assertEqual(isoformat_nosecs(item[0]), '2003-07-24T19:52')

    def test_matches_previous(self):
        item = self.ts.item('2003-07-25 00:00', downwards=True)
        self.assertAlmostEqual(item[1], 108.7)
        self.assertEqual(isoformat_nosecs(item[0]), '2003-07-24T19:52')

    def test_matches_first(self):
        item = self.ts.item('1999-02-28 17:03')
        self.assertAlmostEqual(item[1], 93)
        self.assertEqual(isoformat_nosecs(item[0]), '2003-07-18T18:53')

    def test_matches_last(self):
        item = self.ts.item('2009-02-28 17:03', downwards=True)
        self.assert_(math.isnan(item[1]))
        self.assertEqual(isoformat_nosecs(item[0]), '2005-08-27T00:02')
        self.assertEqual(item[1].flags, set(['DIAMONDS']))

    def test_index_error(self):
        self.assertRaises(IndexError, self.ts.item, '2005-08-27 00:03')
        self.assertRaises(IndexError, self.ts.item, '2003-07-18 18:52',
                                                    downwards=True)


class _Test_Timeseries_min_max_avg_sum(TestCase):

    def setUp(self):
        self.ts = Timeseries(time_step=TimeStep(length_minutes=10))
        self.ts.read(StringIO(big_test_timeseries))

    def test_min(self):
        value = self.ts.min('2004-08-18 00:00', '2004-08-22 00:00')
        self.assertAlmostEqual(value, 28.3)

    def test_max(self):
        value = self.ts.max('2004-08-18 00:00', '2004-08-22 00:00')
        self.assertAlmostEqual(value, 108.7)

    def test_average(self):
        value = self.ts.average('2004-08-18 00:00', '2004-08-22 00:00')
        self.assertAlmostEqual(value, 76.66666666667)

    def test_sum(self):
        value = self.ts.sum('2004-08-18 00:00', '2004-08-22 00:00')
        self.assertAlmostEqual(value, 230.0)

    def test_min_nan(self):
        value = self.ts.min('2004-08-21 00:00', '2004-08-22 12:00')
        self.assert_(math.isnan(value))

    def test_max_nan(self):
        value = self.ts.max('2004-08-21 00:00', '2004-08-22 12:00')
        self.assert_(math.isnan(value))

    def test_average_nan(self):
        value = self.ts.average('2004-08-21 00:00', '2004-08-22 12:00')
        self.assert_(math.isnan(value))

    def test_sum_nan(self):
        value = self.ts.sum('2004-08-21 00:00', '2004-08-22 12:00')
        self.assert_(math.isnan(value))


class _Test_Timeseries_aggregate(TestCase):

    def setUp(self):
        self.ts = Timeseries(time_step=TimeStep(length_minutes=10,
                                                nominal_offset=(0, 0)))
        self.ts.read(StringIO(tenmin_test_timeseries))

    def test_aggregate_hourly_sum(self):
        target_step = TimeStep(length_minutes=60, nominal_offset=(0, 0),
                               interval_type=IntervalType.SUM)
        result, missing = self.ts.aggregate(target_step, missing_allowed=3,
                                            missing_flag="MISS")
        out = StringIO()
        result.write(out)
        self.assertEqual(out.getvalue(), aggregated_hourly_sum)
        out.truncate(0)
        missing.write(out)
        self.assertEqual(out.getvalue(), aggregated_hourly_missing)

    def test_aggregate_hourly_average(self):
        target_step = TimeStep(length_minutes=60, nominal_offset=(0, 0),
                               interval_type=IntervalType.AVERAGE)
        result, missing = self.ts.aggregate(target_step, missing_allowed=3,
                                            missing_flag="MISS")
        correct_ts = Timeseries()
        correct_ts.read(StringIO(aggregated_hourly_average))
        self.assertEqual(len(correct_ts.keys()), len(result.keys()))
        for d in correct_ts.keys():
            self.assertAlmostEqual(result[d], correct_ts[d], 4)
        out = StringIO()
        missing.write(out)
        self.assertEqual(out.getvalue(), aggregated_hourly_missing)

    def test_aggregate_hourly_max(self):
        target_step = TimeStep(length_minutes=60, nominal_offset=(0, 0),
                               interval_type=IntervalType.MAXIMUM)
        result, missing = self.ts.aggregate(target_step, missing_allowed=3,
                                            missing_flag="MISS")
        out = StringIO()
        result.write(out)
        self.assertEqual(out.getvalue(), aggregated_hourly_max)
        out.truncate(0)
        missing.write(out)
        self.assertEqual(out.getvalue(), aggregated_hourly_missing)

    def test_aggregate_hourly_min(self):
        target_step = TimeStep(length_minutes=60, nominal_offset=(0, 0),
                               interval_type=IntervalType.MINIMUM)
        result, missing = self.ts.aggregate(target_step, missing_allowed=3,
                                            missing_flag="MISS")
        out = StringIO()
        result.write(out)
        self.assertEqual(out.getvalue(), aggregated_hourly_min)
        out.truncate(0)
        missing.write(out)
        self.assertEqual(out.getvalue(), aggregated_hourly_missing)


class _Test_Timeseries_vector_aggregate(TestCase):

    def setUp(self):
        self.ts = Timeseries(time_step=TimeStep(length_minutes=10,
                                                nominal_offset=(0, 0)))
        self.ts.read(StringIO(tenmin_vector_test_timeseries))

    def test_aggregate_hourly_vector(self):
        target_step = TimeStep(length_minutes=60, nominal_offset=(0, 0),
                               interval_type=IntervalType.VECTOR_AVERAGE)
        result = (self.ts.aggregate(target_step, missing_allowed=0,
                                    missing_flag="MISS"))[0]
        out = StringIO()
        result.write(out)
        self.assertEqual(out.getvalue(), aggregated_hourly_vector_average)


class _Test_Timeseries_allmiss_aggregate(TestCase):

    def setUp(self):
        self.ts = Timeseries(time_step=TimeStep(length_minutes=10,
                                                nominal_offset=(0, 0)))
        self.ts.read(StringIO(tenmin_allmiss_test_timeseries))

    def test_aggregate_hourly_allmiss(self):
        target_step = TimeStep(length_minutes=60, nominal_offset=(0, 0),
                               interval_type=IntervalType.SUM)
        result = (self.ts.aggregate(target_step, missing_allowed=6,
                                    missing_flag="MISS"))[0]
        out = StringIO()
        result.write(out)
        self.assertEqual(out.getvalue(), aggregated_hourly_allmiss)


events_test_timeseries_1 = textwrap.dedent("""\
                    2008-02-07 09:40,-2.33,
                    2008-02-07 09:50,7.43,
                    2008-02-07 10:00,2.17,
                    2008-02-07 10:10,-0.70,
                    2008-02-07 10:20,-8.32,
                    2008-02-07 10:30,-1.20,
                    2008-02-07 10:40,5.53,
                    2008-02-07 10:50,4.80,
                    2008-02-07 11:00,-8.28,
                    2008-02-07 11:10,3.28,
                    2008-02-07 11:20,-9.56,
                    2008-02-07 11:30,-1.13,
                    2008-02-07 11:40,7.41,
                    2008-02-07 11:50,-0.57,
                    2008-02-07 12:00,5.82,
                    2008-02-07 12:10,-8.92,
                    2008-02-07 12:20,9.19,
                    2008-02-07 12:30,-2.79,
                    2008-02-07 12:40,-5.24,
                    2008-02-07 12:50,-4.87,
                    2008-02-07 13:00,-2.51,
                    2008-02-07 13:10,5.26,
                    """)


events_test_timeseries_2 = textwrap.dedent("""\
                    2008-02-07 09:40,-0.58,
                    2008-02-07 09:50,-6.50,
                    2008-02-07 10:00,-8.14,
                    2008-02-07 10:10,2.78,
                    2008-02-07 10:20,-9.38,
                    2008-02-07 10:30,-5.28,
                    2008-02-07 10:40,7.41,
                    2008-02-07 10:50,3.02,
                    2008-02-07 11:00,5.87,
                    2008-02-07 11:10,0.66,
                    2008-02-07 11:20,-8.72,
                    2008-02-07 11:30,6.05,
                    2008-02-07 11:40,1.65,
                    2008-02-07 11:50,-1.01,
                    2008-02-07 12:00,-6.26,
                    2008-02-07 12:10,-8.62,
                    2008-02-07 12:20,-9.28,
                    2008-02-07 12:30,9.28,
                    2008-02-07 12:40,2.83,
                    2008-02-07 12:50,-0.43,
                    2008-02-07 13:00,5.59,
                    2008-02-07 13:10,-4.87,
                    """)


events_test_timeseries_3 = textwrap.dedent("""\
                    2008-02-07 09:40,4.06,
                    2008-02-07 09:50,2.60,
                    2008-02-07 10:00,5.05,
                    2008-02-07 10:10,-9.76,
                    2008-02-07 10:20,-1.39,
                    2008-02-07 10:30,-9.11,
                    2008-02-07 10:40,-8.25,
                    2008-02-07 10:50,-2.26,
                    2008-02-07 11:00,-7.57,
                    2008-02-07 11:10,1.42,
                    2008-02-07 11:20,-4.30,
                    2008-02-07 11:30,6.54,
                    2008-02-07 11:40,-5.13,
                    2008-02-07 11:50,-6.28,
                    2008-02-07 12:00,8.07,
                    2008-02-07 12:10,-5.70,
                    2008-02-07 12:20,-5.42,
                    2008-02-07 12:30,-1.98,
                    2008-02-07 12:40,5.70,
                    2008-02-07 12:50,4.37,
                    2008-02-07 13:00,-7.03,
                    2008-02-07 13:10,-7.34,
                    """)


class _Test_Timeseries_identify_events(TestCase):

    def setUp(self):
        self.ts1 = Timeseries(time_step=TimeStep(length_minutes=10,
                                                 nominal_offset=(0, 0)))
        self.ts2 = Timeseries(time_step=TimeStep(length_minutes=10,
                                                 nominal_offset=(0, 0)))
        self.ts3 = Timeseries(time_step=TimeStep(length_minutes=10,
                                                 nominal_offset=(0, 0)))
        self.ts1.read(StringIO(events_test_timeseries_1))
        self.ts2.read(StringIO(events_test_timeseries_2))
        self.ts3.read(StringIO(events_test_timeseries_3))

    def test_find_events_more_than_4(self):
        events = identify_events(
            (self.ts1, self.ts2, self.ts3), 4.0, 2, timedelta(minutes=30),
            ntimeseries_end_threshold=1)
        self.assertEqual(len(events), 2)
        self.assertEqual(events[0][0], datetime(2008, 2, 7, 10, 40))
        self.assertEqual(events[0][1], datetime(2008, 2, 7, 11, 0))
        self.assertEqual(events[1][0], datetime(2008, 2, 7, 11, 30))
        self.assertEqual(events[1][1], datetime(2008, 2, 7, 13, 10))

    def test_find_events_less_than_zero(self):
        events = identify_events(
            (self.ts1, self.ts2, self.ts3), 0.0, 3, timedelta(minutes=20),
            ntimeseries_end_threshold=2, reverse=True)
        self.assertEqual(len(events), 4)
        self.assertEqual(events[0][0], datetime(2008, 2, 7, 10, 20))
        self.assertEqual(events[0][1], datetime(2008, 2, 7, 10, 30))
        self.assertEqual(events[1][0], datetime(2008, 2, 7, 11, 20))
        self.assertEqual(events[1][1], datetime(2008, 2, 7, 11, 20))
        self.assertEqual(events[2][0], datetime(2008, 2, 7, 11, 50))
        self.assertEqual(events[2][1], datetime(2008, 2, 7, 11, 50))
        self.assertEqual(events[3][0], datetime(2008, 2, 7, 12, 10))
        self.assertEqual(events[3][1], datetime(2008, 2, 7, 12, 30))

    def test_find_events_in_only_one_timeseries(self):
        events = identify_events((self.ts1,), 4.0, 1, timedelta(minutes=30),
                                 start_date=datetime(2008, 2, 7, 10, 0))
        self.assertEqual(len(events), 3)
        self.assertEqual(events[0][0], datetime(2008, 2, 7, 10, 40))
        self.assertEqual(events[0][1], datetime(2008, 2, 7, 10, 50))
        self.assertEqual(events[1][0], datetime(2008, 2, 7, 11, 40))
        self.assertEqual(events[1][1], datetime(2008, 2, 7, 12, 20))
        self.assertEqual(events[2][0], datetime(2008, 2, 7, 13, 10))
        self.assertEqual(events[2][1], datetime(2008, 2, 7, 13, 10))



#!/usr/bin/env python

from ctypes import CDLL, c_char_p
import platform
import sys
import textwrap

from setuptools import setup, find_packages
from pkg_resources import parse_version


REQUIRED_DICKINSON_VERSION = '0.2.0'


class VersionError(Exception):
    pass


try:
    dickinson = CDLL('dickinson.dll'
                     if platform.system() == 'Windows'
                     else 'libdickinson.so')
    dickinson_version = c_char_p.in_dll(dickinson, 'dickinson_version'
                                        ).value.decode('ascii')
    if dickinson_version != 'dev' and parse_version(dickinson_version) \
            < parse_version(REQUIRED_DICKINSON_VERSION):
        raise VersionError('Too old version of dickinson: {0}'.format(
                           dickinson_version))
except OSError as e:
    sys.stderr.write(str(e) + '\n')
    sys.stderr.write(textwrap.dedent('''
        Please make sure you have installed dickinson
        (see http://dickinson.readthedocs.org/).
        '''))
    sys.exit(1)
except ValueError as e:
    sys.stderr.write(str(e) + '\n')
    sys.stderr.write(textwrap.dedent('''
        Apparently the version of dickinson that is installed doesn't export
        a dickinson_version variable. Maybe it's too old. Please install at
        least version {0} (see http://dickinson.readthedocs.org/).
        '''.format(REQUIRED_DICKINSON_VERSION)))
    sys.exit(1)
except VersionError as e:
    sys.stderr.write(str(e) + '\n')
    sys.stderr.write(textwrap.dedent('''
        Please install at least version {0} of dickinson
        (see http://dickinson.readthedocs.org/)
        '''.format(REQUIRED_DICKINSON_VERSION)))
    sys.exit(1)


kwargs = {
    'name': "pthelma",
    'version': "0.7.0",
    'license': "GPL3",
    'description': "Hydro/meteorological timeseries library",
    'author': "Antonis Christofides",
    'author_email': "anthony@itia.ntua.gr",
    'packages': find_packages(),
    'scripts': ['bin/loggertodb'],
    'test_suite': "tests",
    'install_requires': [
        "pytz",
    ],
}

try:
    # Py2exe stuff; ignored if not in Windows or if py2exe is not installed

    import py2exe
    py2exe  # Does nothing, but lint checkers won't warn about unused py2exe

    import atexit
    import os
    import shutil
    import tempfile
    import urllib2

    # Download and save MSVC++ 2008 redistributable in a temporary directory
    # that will be removed when the program exits
    response = urllib2.urlopen('http://download.microsoft.com/download/'
                               '1/1/1/1116b75a-9ec3-481a-a3c8-1777b5381140/'
                               'vcredist_x86.exe')
    tmpdir = tempfile.mkdtemp()
    atexit.register(lambda: shutil.rmtree(tmpdir, ignore_errors=True))
    vcredist_filename = os.path.join(tmpdir, 'vcredist_x86.exe')
    with open(vcredist_filename, 'wb') as f:
        f.write(response.read())

    # Add the MSVC++ 2008 redistributable and dickinson to the produced files
    kwargs['data_files'] = [('', [r'\Windows\System32\dickinson.dll',
                                  vcredist_filename])]

    # Specify program executable
    kwargs['console'] = ['bin/loggertodb']
except ImportError:
    pass

setup(**kwargs)



#!/usr/bin/python
"""
Timeseries processing
=====================

Copyright (C) 2005-2011 National Technical University of Athens

Copyright (C) 2005 Antonis Christofides

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

import zlib
import random
import re
from datetime import datetime, timedelta
from StringIO import StringIO
from math import sin, cos, atan2, pi
from ConfigParser import ParsingError
from codecs import BOM_UTF8
from os import SEEK_SET

psycopg2 = None  # Do not import unless needed
import math

from ctypes import CDLL, c_int, c_longlong, c_double, c_char, c_char_p, \
    byref, Structure, c_void_p, POINTER, string_at


class T_REC(Structure):
    _fields_ = [("timestamp", c_longlong),
                ("null", c_int),
                ("value", c_double),
                ("flags", c_char_p)]


class T_INTERVAL(Structure):
    _fields_ = [("start_date", c_longlong),
                ("end_date", c_longlong)]


class T_INTERVALLIST(Structure):
    _fields_ = [("intervals", POINTER(T_INTERVAL)),
                ("n", c_int)]


class T_TIMESERIESLIST(Structure):
    _fields = [("ts", c_void_p),
               ("n", c_int)]

import platform
dickinson = CDLL('dickinson.dll'
                 if platform.system() == 'Windows'
                 else 'libdickinson.so')

dickinson.ts_get_item.restype = T_REC
dickinson.ts_create.restype = c_void_p
dickinson.tsl_create.restype = POINTER(T_TIMESERIESLIST)
dickinson.il_create.restype = POINTER(T_INTERVALLIST)
dickinson.ts_min.restype = c_double
dickinson.ts_max.restype = c_double
dickinson.ts_average.restype = c_double
dickinson.ts_sum.restype = c_double
dickinson.ts_write.restype = POINTER(c_char)

re_compiled = re.compile(r'''^(\d{4})-(\d{1,2})-(\d{1,2})
                         (?: [ tT] (\d{1,2}):(\d{1,2}) )?''',
                         re.VERBOSE)


def datetime_from_iso(isostring):
    m = re_compiled.match(isostring)
    if m is None:
        raise ValueError('%s is not a valid date and time' % (isostring))
    args = []
    for x in m.groups():
        if not (x is None):
            args.append(int(x))
    return datetime(*args)


def isoformat_nosecs(adatetime, sep='T'):
    return adatetime.isoformat(sep)[:16]


_DT_BASE = datetime(1970, 1, 1, 0, 0)


def _datetime_to_time_t(d):
    """Convert d (a datetime or iso string) to number of seconds since 1970"""
    if not isinstance(d, datetime):
        d = datetime_from_iso(d)
    delta = d - _DT_BASE
    return delta.days * 86400L + delta.seconds


def _time_t_to_datetime(t):
    """Convert t (number of seconds since 1970) to datetime."""
    return _DT_BASE + timedelta(t / 86400, t % 86400)


def read_timeseries_tail_from_db(db, id):
    c = db.cursor()
    c.execute("""SELECT bottom FROM ts_records
                 WHERE id=%d""" % (id,))
    r = c.fetchone()
    bottom_lines = r[0].strip().splitlines()
    c.close()
    if not bottom_lines:
        raise ValueError('No time series with id=%d found' % (id,))
    return bottom_lines[-1].split(',')[:2]


def timeseries_bounding_dates_from_db(db, id):
    c = db.cursor()
    c.execute("""SELECT timeseries_start_date(%d)""" % (id,))
    r = c.fetchone()
    start_date = r[0]
    c.execute("""SELECT timeseries_end_date(%d)""" % (id,))
    r = c.fetchone()
    end_date = r[0]
    return start_date, end_date


class IntervalType:
    SUM = 1
    AVERAGE = 2
    MINIMUM = 3
    MAXIMUM = 4
    VECTOR_AVERAGE = 5


class TimeStep:

    def __init__(self, length_minutes=0, length_months=0, interval_type=None,
                 nominal_offset=None, actual_offset=(0, 0)):
        self.length_minutes = length_minutes
        self.length_months = length_months
        self.nominal_offset = nominal_offset
        self.actual_offset = actual_offset
        self.interval_type = interval_type

    def _check_nominal_offset(self):
        """Called whenever an operation requires a nominal offset; verifies
        that the nominal offset is not None, otherwise raises exception."""
        if self.nominal_offset:
            return
        raise ValueError("This operation requires a nominal offset")

    def up(self, timestamp):
        self._check_nominal_offset()
        if self.length_minutes:
            required_modulo = self.nominal_offset[0]
            if required_modulo < 0:
                required_modulo += self.length_minutes
            reference_date = timestamp.replace(day=1, hour=0, minute=0)
            d = timestamp - reference_date
            diff_in_minutes = d.days * 1440 + d.seconds / 60
            actual_modulo = diff_in_minutes % self.length_minutes
            result = timestamp - timedelta(
                minutes=actual_modulo - required_modulo)
            while result < timestamp:
                result += timedelta(minutes=self.length_minutes)
            return result
        else:
            y = timestamp.year - 1
            m = 1 + self.nominal_offset[1]
            result = timestamp.replace(
                year=y, month=m, day=1, hour=0, minute=0) + \
                timedelta(minutes=self.nominal_offset[0])
            while result < timestamp:
                m += self.length_months
                if m > 12:
                    m -= 12
                    y += 1
                result = timestamp.replace(
                    year=y, month=m, day=1, hour=0, minute=0) + \
                    timedelta(minutes=self.nominal_offset[0])
            return result

    def down(self, timestamp):
        self._check_nominal_offset()
        if self.length_minutes:
            required_modulo = self.nominal_offset[0]
            if required_modulo < 0:
                required_modulo += self.length_minutes
            reference_date = timestamp.replace(day=1, hour=0, minute=0)
            d = timestamp - reference_date
            diff_in_minutes = d.days * 1440 + d.seconds / 60
            actual_modulo = diff_in_minutes % self.length_minutes
            result = timestamp + timedelta(minutes=required_modulo -
                                           actual_modulo)
            while result > timestamp:
                result -= timedelta(minutes=self.length_minutes)
        elif self.length_months:
            y = timestamp.year + 1
            m = 1 + self.nominal_offset[1]
            result = timestamp.replace(
                year=y, month=m, day=1, hour=0, minute=0) + \
                timedelta(minutes=self.nominal_offset[0])
            while result > timestamp:
                m -= self.length_months
                if m < 1:
                    m += 12
                    y -= 1
                result = timestamp.replace(
                    year=y, month=m, day=1, hour=0, minute=0) + \
                    timedelta(minutes=self.nominal_offset[0])
        else:
            assert(False)
        return result

    def next(self, timestamp):
        timestamp = self.up(timestamp)
        m = timestamp.month
        y = timestamp.year
        m += self.length_months
        while m > 12:
            m -= 12
            y += 1
        timestamp = timestamp.replace(year=y, month=m)
        return timestamp + timedelta(minutes=self.length_minutes)

    def previous(self, timestamp):
        timestamp = self.down(timestamp)
        m = timestamp.month
        y = timestamp.year
        m -= self.length_months
        while m < 1:
            m += 12
            y -= 1
        timestamp = timestamp.replace(year=y, month=m)
        return timestamp - timedelta(minutes=self.length_minutes)

    def actual_timestamp(self, timestamp):
        m = timestamp.month + self.actual_offset[1]
        y = timestamp.year
        while m > 12:
            m -= 12
            y += 1
        while m < 1:
            m += 12
            y -= 1
        return timestamp.replace(year=y, month=m) + \
            timedelta(minutes=self.actual_offset[0])

    def containing_interval(self, timestamp):
        result = self.down(timestamp)
        while self.actual_timestamp(result) >= timestamp:
            result = self.previous(result)
        while self.actual_timestamp(result) < timestamp:
            result = self.next(result)
        return result

    def interval_endpoints(self, nominal_timestamp):
        end_date = self.actual_timestamp(nominal_timestamp)
        start_date = self.actual_timestamp(self.previous(nominal_timestamp))
        return start_date, end_date


class _Tsvalue(float):

    def __new__(cls, value, flags=[]):
        return super(_Tsvalue, cls).__new__(cls, value)

    def __init__(self, value, flags=[]):
        self.flags = set(flags)


def strip_trailing_zeros(s):
    if s.rfind('.') > 0:
        s = s.rstrip('0')
    if s[-1] == '.':
        s = s[:-1]
    return s


class Timeseries(dict):

    # Some constants for how timeseries records are distributed in
    # top, middle, and bottom.  Records are appended to bottom, except
    # if bottom would then have more than MAX_BOTTOM plus/minus
    # MAX_BOTTOM_NOISE records (i.e. random noise, evenly distributed
    # between -MAX_BOTTOM_NOISE and +MAX_BOTTOM_NOISE is added to
    # MAX_BOTTOM; this is to avoid reaching circumstances where 20
    # timeseries will be repacked altogether).  If a timeseries is
    # stored entirely from scratch, then all records go to bottom if
    # they are less than MAX_ALL_BOTTOM; otherwise ROWS_IN_TOP_BOTTOM
    # go to top, another as much go to bottom, the rest goes to
    # middle.
    MAX_BOTTOM = 1000
    MAX_BOTTOM_NOISE = 150
    MAX_ALL_BOTTOM = 80
    ROWS_IN_TOP_BOTTOM = 20

    SQLDRIVER_NONE = 0
    SQLDRIVER_PSYCOPG2 = 1

    def __init__(self, id=0, time_step=None, unit=u'', title=u'', timezone=u'',
                 variable=u'', precision=None, comment=u'',
                 driver=SQLDRIVER_PSYCOPG2):
        self.ts_handle = None
        self.id = id
        if time_step:
            self.time_step = time_step
        else:
            self.time_step = TimeStep()
        self.unit = unit
        self.title = title
        self.timezone = timezone
        self.variable = variable
        self.precision = precision
        self.comment = comment
        assert(driver in (self.SQLDRIVER_NONE, self.SQLDRIVER_PSYCOPG2))
        self.driver = driver
        self.ts_handle = c_void_p(dickinson.ts_create())
        if self.ts_handle.value == 0:
            raise MemoryError.Create('Could not allocate memory '
                                     'for time series object.')
        # When the object is being destroyed, it doesn't always have access
        # to module globals (see Python documentation on the __del__ special
        # method), and therefore we keep a copy of the required globals here.
        self.__dickinson = dickinson

    def __del__(self):
        if self.ts_handle is None:
            return
        if self.ts_handle.value != 0:
            self.__dickinson.ts_free(self.ts_handle)
        self.ts_handle.value = 0

    def __len__(self):
        return dickinson.ts_length(self.ts_handle)

    def __delitem__(self, key):
        index_c = dickinson.ts_get_i(self.ts_handle,
                                     c_longlong(_datetime_to_time_t(key)))
        if index_c < 0:
            raise KeyError(
                'No such record: ' + (isoformat_nosecs(key, ' ')
                if isinstance(key, datetime)
                else key))
        dickinson.ts_delete_item(self.ts_handle, index_c)

    def __contains__(self, key):
        index_c = dickinson.ts_get_i(self.ts_handle,
                                     c_longlong(_datetime_to_time_t(key)))
        if index_c < 0:
            return False
        else:
            return True

    def __setitem__(self, key, value):
        timestamp_c = c_longlong(_datetime_to_time_t(key))
        index_c = dickinson.ts_get_i(self.ts_handle, timestamp_c)
        if isinstance(value, _Tsvalue):
            tsvalue = value
        elif isinstance(value, tuple):
            tsvalue = _Tsvalue(value[0], value[1])
        elif index_c >= 0:
            tsvalue = _Tsvalue(value, self[key].flags)
        else:
            tsvalue = _Tsvalue(value, [])
        if math.isnan(tsvalue):
            null_c = 1
            value_c = c_double(0)
        else:
            null_c = 0
            value_c = c_double(tsvalue)
        flags_c = c_char_p(' '.join(tsvalue.flags))
        err_str_c = c_char_p()
        index_c = c_int()
        err_no_c = dickinson.ts_insert_record(
            self.ts_handle, timestamp_c, null_c, value_c, flags_c, c_int(1),
            byref(index_c), byref(err_str_c))
        if err_no_c != 0:
            raise Exception('Something wrong occured in dickinson '
                            'function when setting a time series value. '
                            'Error message: ' + repr(err_str_c.value))

    def __getitem__(self, key):
        if not isinstance(key, datetime) and (
                (not isinstance(key, unicode) and
                 not isinstance(key, str))
                or len(key) < 4 or not key[0].isdigit()):
            raise KeyError(key)
        timestamp_c = c_longlong(_datetime_to_time_t(key))
        index_c = dickinson.ts_get_i(self.ts_handle, timestamp_c)
        if index_c < 0:
            raise KeyError(
                'No such record: ' + (isoformat_nosecs(key, ' ')
                if isinstance(key, datetime)
                else key))
        arec = dickinson.ts_get_item(self.ts_handle, index_c)
        if arec.null == 1:
            value = float('NaN')
        else:
            value = arec.value
        flags = arec.flags
        flags = flags.split()
        return _Tsvalue(value, flags)

    def get(self, key, default=None):
        if self.__contains__(key):
            return self.__getitem__(key)
        else:
            return default

    def keys(self):
        a = []
        i = 0
        while i < dickinson.ts_length(self.ts_handle):
            rec = dickinson.ts_get_item(self.ts_handle, c_int(i))
            a.append(_time_t_to_datetime(rec.timestamp))
            i += 1
        return a

    def iterkeys(self):
        i = 0
        while i < dickinson.ts_length(self.ts_handle):
            rec = dickinson.ts_get_item(self.ts_handle, c_int(i))
            yield _time_t_to_datetime(rec.timestamp)
            i += 1
    __iter__ = iterkeys

    def clear(self):
        i = dickinson.ts_length(self.ts_handle)
        while i >= 0:
            dickinson.ts_delete_item(self.ts_handle, i)
            i -= 1

    def read(self, fp, line_number=1):
        err_str_c = c_char_p()
        errline = c_int()
        try:
            if dickinson.ts_readfromstring(
                    c_char_p(fp.read()), self.ts_handle, byref(errline),
                    byref(err_str_c)):
                line_number += errline.value - 1
                raise ValueError('Error when reading time series '
                                 'line from I/O: ' + repr(err_str_c.value))
            line_number += errline.value - 1
        except Exception, e:
            e.args = e.args + (line_number,)
            raise

    def write(self, fp, start=None, end=None):
        errstr = c_char_p()
        start_date = c_longlong.in_dll(dickinson, "LONG_TIME_T_MIN") \
            if start is None else c_longlong(_datetime_to_time_t(start))
        end_date = c_longlong.in_dll(dickinson, "LONG_TIME_T_MAX") \
            if end is None else c_longlong(_datetime_to_time_t(end))
        text = dickinson.ts_write(
            self.ts_handle,
            c_int(self.precision if self.precision is not None else -9999),
            start_date, end_date, byref(errstr))
        if not text:
            if not errstr:
                fp.write('')
                return
            raise IOError('Error when writing time series: %s' %
                          (errstr.value,))
        try:
            fp.write(string_at(text))
        finally:
            dickinson.freemem(text)

    def write_plain_values(self, fp, nullstr=''):
        i = 0
        lines = []
        while i < dickinson.ts_length(self.ts_handle):
            rec = dickinson.ts_get_item(self.ts_handle, c_int(i))
            lines.append(nullstr if rec.null else str(rec.value))
            i += 1
        fp.write('\r\n'.join(lines))

    def delete_from_db(self, db):
        c = db.cursor()
        c.execute("""DELETE FROM ts_records
                     WHERE id=%d""" % (self.id))
        self.clear()
        c.close()

    def __read_meta_line(self, fp):
        """Read one line from a file format header and return a (name, value)
        tuple, where name is lowercased. Returns ('', '') if the next line is
        blank. Raises ParsingError if next line in fp is not a valid header
        line."""
        line = fp.readline()
        (name, value) = '', ''
        if line.isspace():
            return (name, value)
        if line.find('=') > 0:
            (name, value) = line.split('=', 1)
            name = name.rstrip().lower()
            value = value.strip()
        for c in name:
            if c.isspace():
                name = ''
                break
        if not name:
            raise ParsingError(("Invalid file header line"))
        return (name, value)

    def __read_meta(self, fp):
        """Read the headers of a file in file format into the instance
        attributes and return the line number of the first data line of the
        file.
        """
        def read_minutes_months(s):
            """Return a (minutes, months) tuple after parsing a "M,N" string.
            """
            try:
                (minutes, months) = [int(x.strip()) for x in s.split(',')]
                return minutes, months
            except Exception:
                raise ParsingError(('Value should be "minutes, months"'))
        #Ignore the BOM_UTF8 byte mark if present by advancing
        saved_pos = fp.tell()
        if fp.read(len(BOM_UTF8)) != BOM_UTF8:
            fp.seek(saved_pos)
        line_number = 1
        try:
# Changed the behavior when 'version=2' is not present..
# If version=2 not parsed, then the file is considered as
# a plain text file with time series data only, withoud
# head section with meta information. Stefanos, 2011-06-10
            try:
                saved_pos = fp.tell()
                (name, value) = self.__read_meta_line(fp)
                if name != 'version' or value != '2':
                    raise ParsingError(("The first line must be Version=2"))
            except ParsingError:
                fp.seek(saved_pos, SEEK_SET)
                return line_number
            line_number += 1
            (name, value) = self.__read_meta_line(fp)
            while name:
                if name == 'unit':
                    self.unit = value
                elif name == 'title':
                    self.title = value
                elif name == 'timezone':
                    self.timezone = value
                elif name == 'variable':
                    self.variable = value
                elif name == 'time_step':
                    minutes, months = read_minutes_months(value)
                    self.time_step.length_minutes = minutes
                    self.time_step.length_months = months
                elif name == 'nominal_offset':
                    self.time_step.nominal_offset = read_minutes_months(value)
                elif name == 'actual_offset':
                    self.time_step.actual_offset = read_minutes_months(value)
                elif name == 'interval_type':
                    it = IntervalType
                    v = value.lower()
                    if v == 'sum':
                        self.time_step.interval_type = it.SUM
                    elif v == 'average':
                        self.time_step.interval_type = it.AVERAGE
                    elif v == 'maximum':
                        self.time_step.interval_type = it.MAXIMUM
                    elif v == 'minimum':
                        self.time_step.interval_type = it.MINIMUM
                    elif v == 'vector_average':
                        self.time_step.interval_type = it.VECTOR_AVERAGE
                    elif v == '':
                        self.time_step.interval_type = None
                    else:
                        raise ParsingError(("Invalid interval type"))
                elif name == 'precision':
                    try:
                        self.precision = int(value)
                    except TypeError as e:
                        raise ParsingError(e.args)
                elif name == 'comment':
                    if self.comment:
                        self.comment += '\n'
                    self.comment += value.decode('utf-8')
                elif name == 'count':
                    pass
                line_number += 1
                (name, value) = self.__read_meta_line(fp)
                if not name and not value:
                    return line_number
            return line_number
        except ParsingError, e:
            e.args = e.args + (line_number,)
            raise

    def read_file(self, fp):
        line_number = self.__read_meta(fp)
        self.read(fp, line_number=line_number)

    def write_file(self, fp):
        fp.write(u"Version=2\r\n")
        if self.unit:
            fp.write(u"Unit=%s\r\n" % (self.unit,))
        fp.write(u"Count=%d\r\n" % (len(self),))
        if self.title:
            fp.write(u"Title=%s\r\n" % (self.title,))
        for line in self.comment.splitlines():
            fp.write(u"Comment=%s\r\n" % (line,))
        if self.timezone:
            fp.write(u"Timezone=%s\r\n" % (self.timezone,))
        if self.time_step.length_minutes or self.time_step.length_months:
            fp.write(u"Time_step=%d,%d\r\n" % (self.time_step.length_minutes,
                                               self.time_step.length_months))
            if self.time_step.nominal_offset:
                fp.write(u"Nominal_offset=%d,%d\r\n" %
                         self.time_step.nominal_offset)

            fp.write(u"Actual_offset=%d,%d\r\n" %
                     self.time_step.actual_offset)
        if self.time_step.interval_type:
            fp.write(
                u"Interval_type=%s\r\n" % ({
                IntervalType.SUM: u"sum", IntervalType.AVERAGE: u"average",
                IntervalType.MAXIMUM: u"maximum",
                IntervalType.MINIMUM: u"minimum",
                IntervalType.VECTOR_AVERAGE: u"vector_average"
                }[self.time_step.interval_type],))
        if self.variable:
            fp.write(u"Variable=%s\r\n" % (self.variable,))
        if self.precision is not None:
            fp.write(u"Precision=%d\r\n" % (self.precision,))

        fp.write("\r\n")
        self.write(fp)

    def read_from_db(self, db, bottom_only=False):
        c = db.cursor()
        if not bottom_only:
            querystr = """SELECT top, middle, bottom FROM ts_records
                          WHERE id=%d"""
        else:
            querystr = """SELECT bottom FROM ts_records WHERE id=%d"""
        c.execute(querystr % (self.id))
        r = c.fetchone()
        self.clear()
        if r:
            if bottom_only:
                (bottom,) = r
            else:
                (top, middle, bottom) = r
            if not bottom_only and top:
                self.read(StringIO(top))
                self.read(StringIO(zlib.decompress(middle)))
            self.read(StringIO(bottom))
        c.close()

    def blob_create(self, s):
        if self.driver == self.SQLDRIVER_NONE:
            return ''
        elif self.driver == self.SQLDRIVER_PSYCOPG2:
            global psycopg2
            if psycopg2 is None:
                import psycopg2
            return psycopg2.Binary(s)
        else:
            assert(False)

    def write_to_db(self, db, transaction=None, commit=True):
        if transaction is None:
            transaction = db
        fp = StringIO()
        if len(self) < Timeseries.MAX_ALL_BOTTOM:
            top = ''
            middle = None
            self.write(fp)
            bottom = fp.getvalue()
        else:
            dates = sorted(self.keys())
            self.write(fp, end=dates[Timeseries.ROWS_IN_TOP_BOTTOM - 1])
            top = fp.getvalue()
            fp.truncate(0)
            self.write(fp, start=dates[Timeseries.ROWS_IN_TOP_BOTTOM],
                       end=dates[-(Timeseries.ROWS_IN_TOP_BOTTOM + 1)])
            middle = self.blob_create(zlib.compress(fp.getvalue()))
            fp.truncate(0)
            self.write(fp, start=dates[-Timeseries.ROWS_IN_TOP_BOTTOM])
            bottom = fp.getvalue()
        fp.close()
        c = db.cursor()
        c.execute("DELETE FROM ts_records WHERE id=%d" % (self.id))
        c.execute("""INSERT INTO ts_records (id, top, middle, bottom)
                     VALUES (%s, %s, %s, %s)""", (self.id, top, middle,
                  bottom))
        c.close()
        if commit:
            transaction.commit()

    def append_to_db(self, db, transaction=None, commit=True):
        """Append the contained records to the timeseries stored in the
        database."""
        if transaction is None:
            transaction = db
        if not len(self):
            return
        c = db.cursor()
        bottom_ts = Timeseries()
        c.execute("SELECT bottom FROM ts_records WHERE id=%d" %
                  (self.id))
        r = c.fetchone()
        if r:
            bottom_ts.read(StringIO(r[0]))
            if bottom_ts.keys() and max(bottom_ts.keys()) >= min(self.keys()):
                raise ValueError((
                    "Cannot append time series: "
                    "its first record (%s) has a date earlier than the last "
                    "record (%s) of the timeseries to append to.")
                    % (str(min(self.keys())), str(max(bottom_ts.keys()))))
        max_bottom = Timeseries.MAX_BOTTOM + random.randrange(
            -Timeseries.MAX_BOTTOM_NOISE, Timeseries.MAX_BOTTOM_NOISE)
        if len(bottom_ts) and len(bottom_ts) + len(self) < max_bottom:
            fp = StringIO()
            bottom_ts.write(fp)
            self.write(fp)
            c.execute("""UPDATE ts_records SET bottom=%s
                         WHERE id=%s""", (fp.getvalue(), self.id))
            fp.close()
            if commit:
                transaction.commit()
        else:
            ts = Timeseries(self.id)
            ts.read_from_db(db)
            ts.append(self)
            ts.write_to_db(db, transaction=transaction, commit=commit)
        c.close()

    def append(self, b):
        if len(self) and len(b) and max(self.keys()) >= min(b.keys()):
            raise ValueError(
                "Cannot append: the first record (%s) of the "
                "time series to append has a date earlier than the last "
                "record (%s) of the timeseries to append to."
                % (str(min(b.keys())), str(max(self.keys()))))
        err_str_c = c_char_p()
        if dickinson.ts_merge(self.ts_handle, b.ts_handle, byref(err_str_c)) \
                != 0:
            raise Exception('An exception has occured when trying to '
                            'merge time series. Error message: ' +
                            repr(err_str_c.value))

    def bounding_dates(self):
        if len(self):
            rec1 = dickinson.ts_get_item(self.ts_handle, c_int(0))
            rec2 = dickinson.ts_get_item(self.ts_handle, c_int(len(self) - 1))
            return _time_t_to_datetime(rec1.timestamp),\
                _time_t_to_datetime(rec2.timestamp)
        else:
            return None

    def items(self, pos=None):
        a = []
        i = 0
        if pos is not None:
            if pos < 0 or pos >= dickinson.ts_length(self.ts_handle):
                raise IndexError("Index (%d) out of "
                                 "bounds (%d, %d)" % (pos, 0,
                                 dickinson.ts_length(self.ts_handle) - 1,))
            i = pos
        while i < dickinson.ts_length(self.ts_handle):
            rec = dickinson.ts_get_item(self.ts_handle, c_int(i))
            a.append((_time_t_to_datetime(rec.timestamp),
                     _Tsvalue(float('NaN') if rec.null else rec.value,
                              rec.flags.split())))
            if pos is not None:
                break
            i += 1
        return a if pos is None else a[0]

    def index(self, date, downwards=False):
        timestamp_c = c_longlong(_datetime_to_time_t(date))
        if not downwards:
            pos = dickinson.ts_get_next_i(self.ts_handle, timestamp_c)
        else:
            pos = dickinson.ts_get_prev_i(self.ts_handle, timestamp_c)
        if pos < 0:
            raise IndexError(
                "There is no item in the timeseries on or %s %s"
                % ("before" if downwards else "after", str(date)))
        return pos

    def item(self, date, downwards=False):
        rec = dickinson.ts_get_item(self.ts_handle, c_int(self.index(date,
                                    downwards)))
        return (
            _time_t_to_datetime(rec.timestamp),
            _Tsvalue(float('NaN') if rec.null else rec.value,
                     rec.flags.split()))

    def _get_bounding_indexes(self, start_date, end_date):
        """Return a tuple, (start_index, end_index).  If arguments are None,
        the respective bounding date is considered. The results are the start
        and end indexes in items() of all items that are in the specified
        interval.
        """
        (s, e) = self.bounding_dates()
        if not start_date:
            start_date = s
        if not end_date:
            end_date = e
        return (self.index(start_date), self.index(end_date, downwards=True))

    def min(self, start_date=None, end_date=None):
        start_date = c_longlong.in_dll(dickinson, "LONG_TIME_T_MIN") \
            if start_date is None \
            else c_longlong(_datetime_to_time_t(start_date))
        end_date = c_longlong.in_dll(dickinson, "LONG_TIME_T_MAX") \
            if end_date is None else c_longlong(_datetime_to_time_t(end_date))
        return dickinson.ts_min(self.ts_handle, start_date, end_date)

    def max(self, start_date=None, end_date=None):
        start_date = c_longlong.in_dll(dickinson, "LONG_TIME_T_MIN") \
            if start_date is None \
            else c_longlong(_datetime_to_time_t(start_date))
        end_date = c_longlong.in_dll(dickinson, "LONG_TIME_T_MAX") \
            if end_date is None else c_longlong(_datetime_to_time_t(end_date))
        return dickinson.ts_max(self.ts_handle, start_date, end_date)

    def average(self, start_date=None, end_date=None):
        start_date = c_longlong.in_dll(dickinson, "LONG_TIME_T_MIN") \
            if start_date is None \
            else c_longlong(_datetime_to_time_t(start_date))
        end_date = c_longlong.in_dll(dickinson, "LONG_TIME_T_MAX") \
            if end_date is None else c_longlong(_datetime_to_time_t(end_date))
        return dickinson.ts_average(self.ts_handle, start_date, end_date)

    def sum(self, start_date=None, end_date=None):
        start_date = c_longlong.in_dll(dickinson, "LONG_TIME_T_MIN") \
            if start_date is None \
            else c_longlong(_datetime_to_time_t(start_date))
        end_date = c_longlong.in_dll(dickinson, "LONG_TIME_T_MAX") \
            if end_date is None else c_longlong(_datetime_to_time_t(end_date))
        return dickinson.ts_sum(self.ts_handle, start_date, end_date)

    def aggregate(self, target_step, missing_allowed=0.0, missing_flag="",
                  last_incomplete=False, all_incomplete=False):

        def aggregate_one_step(d, test_run=False, explicit_components=0):
            """Return tuple of ((result value, flags), missing) for a single
            target stamp d."""

            def timedeltadivide(a, b):
                """Divide timedelta a by timedelta b."""
                a = a.days * 86400 + a.seconds
                b = b.days * 86400 + b.seconds
                return a / b

            d_start_date, d_end_date = target_step.interval_endpoints(d)
            start_nominal = self.time_step.containing_interval(d_start_date)
            end_nominal = self.time_step.containing_interval(d_end_date)
            s = start_nominal
            it = target_step.interval_type
            if it in (IntervalType.SUM, IntervalType.AVERAGE):
                aggregate_value = 0.0
            elif it == IntervalType.MAXIMUM:
                aggregate_value = -1e38
            elif it == IntervalType.MINIMUM:
                aggregate_value = 1e38
            elif it == IntervalType.VECTOR_AVERAGE:
                aggregate_value = (0, 0)
            elif it is None:
                aggregate_value = float('NaN')
            else:
                assert(False)
            missing = 0.0
            total_components = 0.0
            divider = 0.0
            source_has_missing = False
            while s <= end_nominal:
                s_start_date, s_end_date = self.time_step.interval_endpoints(s)
                used_interval = s_end_date - s_start_date
                unused_interval = timedelta()
                if s_start_date < d_start_date:
                    out = d_start_date - s_start_date
                    unused_interval += out
                    used_interval -= out
                if s_end_date > d_end_date:
                    out = s_end_date - d_end_date
                    unused_interval += out
                    used_interval -= out
                pct_used = timedeltadivide(used_interval,
                                           unused_interval + used_interval)
                total_components += pct_used
                if math.isnan(self.get(s, float('NaN'))):
                    missing += pct_used
                    if last_incomplete or all_incomplete:
                        if s > self.bounding_dates()[1]:
                            missing -= pct_used
                            if test_run:
                                explicit_components = \
                                    total_components - pct_used
                                break
                    s = self.time_step.next(s)
                    continue
                divider += pct_used
                if missing_flag in self[s].flags:
                    source_has_missing = True
                if it in (IntervalType.SUM, IntervalType.AVERAGE):
                    aggregate_value += self.get(s, 0) * pct_used
                elif it == IntervalType.MAXIMUM:
                    if pct_used > 0:
                        aggregate_value = max(aggregate_value, self[s])
                elif it == IntervalType.MINIMUM:
                    if pct_used > 0:
                        aggregate_value = min(aggregate_value, self[s])
                elif it == IntervalType.VECTOR_AVERAGE:
                    aggregate_value = (
                        aggregate_value[0] + cos(self[s] / 180 * pi)
                        * pct_used,
                        aggregate_value[1] + sin(self[s] / 180 * pi)
                        * pct_used)
                elif it is None:
                    total_components, missing = 1, 1
                    aggregate_value = self.get(end_nominal, float('NaN'))
                    if not math.isnan(aggregate_value):
                        missing = 0
                    break
                else:
                    assert(False)
                if all_incomplete and not test_run:
                    if total_components >= explicit_components:
                        break
                s = self.time_step.next(s)
            flag = []
            if not test_run and missing / total_components > \
                    missing_allowed / total_components + 1e-5 \
                    or abs(missing - total_components) < 1e-36:
                aggregate_value = float('NaN')
            else:
                if (missing / total_components > 1e-36) or \
                        source_has_missing:
                    flag = [missing_flag]
                if it == IntervalType.AVERAGE:
                    aggregate_value /= divider
                elif it == IntervalType.VECTOR_AVERAGE:
                    aggregate_value = atan2(aggregate_value[1],
                                            aggregate_value[0]) / pi * 180
                    while aggregate_value < 0:
                        aggregate_value += 360
                    if abs(aggregate_value - 360) < 1e-7:
                        aggregate_value = 0
            return (aggregate_value, flag), missing, explicit_components

        source_start_date, source_end_date = self.bounding_dates()
        target_start_date = target_step.previous(source_start_date)
        target_end_date = target_step.next(source_end_date)
        result = Timeseries(time_step=target_step)
        missing = Timeseries(time_step=target_step)
        ec = 0
        it = 0

        if all_incomplete:
            d = target_end_date
            while d >= target_start_date:
                dummy1, dummy2, ec = aggregate_one_step(
                    d, test_run=True, explicit_components=ec)
                it += 1
                if ec > 0 or it > 3:
                    break
                d = target_step.previous(d)
            if ec == 0:
                ec = 1e9

        d = target_start_date
        while d <= target_end_date:
            result[d], missing[d], dummy3 = aggregate_one_step(
                d, test_run=False, explicit_components=ec)
            d = target_step.next(d)
        while math.isnan(result.get(target_start_date, 0)):
            del result[target_start_date]
            del missing[target_start_date]
            target_start_date = target_step.next(target_start_date)
        while math.isnan(result.get(target_end_date, 0)):
            del result[target_end_date]
            del missing[target_end_date]
            target_end_date = target_step.previous(target_end_date)
        return result, missing


def identify_events(ts_list,
                    start_threshold, ntimeseries_start_threshold,
                    time_separator,
                    end_threshold=None, ntimeseries_end_threshold=None,
                    start_date=None, end_date=None, reverse=False):
    if end_threshold is None:
        end_threshold = start_threshold
    if ntimeseries_end_threshold is None:
        ntimeseries_end_threshold = ntimeseries_start_threshold
    range_start_date = c_longlong.in_dll(dickinson, "LONG_TIME_T_MIN") \
        if start_date is None else c_longlong(_datetime_to_time_t(start_date))
    range_end_date = c_longlong.in_dll(dickinson, "LONG_TIME_T_MAX") \
        if end_date is None else c_longlong(_datetime_to_time_t(end_date))
    search_range = T_INTERVAL(range_start_date, range_end_date)
    try:
        a_timeseries_list = dickinson.tsl_create()
        a_interval_list = dickinson.il_create()
        if (not a_timeseries_list) or (not a_interval_list):
            raise MemoryError('Insufficient memory')
        for t in ts_list:
            if dickinson.tsl_append(a_timeseries_list, t.ts_handle):
                raise MemoryError('Insufficient memory')
        errstr = c_char_p()
        if dickinson.ts_identify_events(
                a_timeseries_list, search_range, c_int(reverse),
                c_double(start_threshold), c_double(end_threshold),
                c_int(ntimeseries_start_threshold),
                c_int(ntimeseries_end_threshold),
                c_longlong(time_separator.days * 86400L +
                           time_separator.seconds),
                a_interval_list, byref(errstr)):
            raise Exception(errstr.value)
        result = []
        for i in range(a_interval_list.contents.n):
            a_interval = a_interval_list.contents.intervals[i]
            result.append((_time_t_to_datetime(a_interval.start_date),
                           _time_t_to_datetime(a_interval.end_date)))
        return result
    finally:
        dickinson.il_free(a_interval_list)
        dickinson.tsl_free(a_timeseries_list)
